'''
Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
Created on Dec 14, 2017
@author: Niels Lubbes
'''

from ns_lattice.sage_interface import sage_identity_matrix

from ns_lattice.class_div import Div

from ns_lattice.class_dp_lattice import DPLattice

from ns_lattice.class_ns_tools import NSTools


def break_col( col, max_len, row_num ):
    '''
    Splits a string by commas in to a list of "row_num" substrings.
    The length of each substring should not exceed "max_len", if
    this is possible by splitting via commas.
    
    Parameters
    ----------
    col : int
    max_len : int
    row_num : int
    
    Returns
    -------
    list<string>        
    '''

    col = str( col )

    col_lst = []
    if ',' not in col or len( col ) < max_len:
        col_lst = [ str( col ) ]
    else:
        s_lst = str( col ).split( ',' )
        ts = ''
        for s in s_lst:

            if len( ts + s + ', ' ) >= max_len:
                col_lst += [ts]
                ts = ''

            ts += s + ', '
            if ts[-3:] in ['), ', '], ']:
                ts = ts[:-2]  # remove the comma at the end

        # add remainder
        col_lst += [ts]

    # we cannot break up a column in too many rows
    if row_num - len( col_lst ) < 0:
        raise Exception( 'Parameter row_num is too small: ', row_num, '<', len( col_lst ) )

    # return row_num columns
    for i in range( row_num - len( col_lst ) ):
        col_lst += ['']

    return col_lst


def refine_table( table, max_len = 50, row_num = 5 ):
    '''
    Break a long row into several short rows
    
    Parameters
    ----------
    row : list<list<object>>
    
    Returns
    -------
    list<list<object>>    
    '''
    new_table = []
    for ri in range( len( table ) ):
        for i in range( row_num ):

            new_row = []
            for ci in range( len( table[ri] ) ):
                new_row += [ break_col( table[ri][ci], max_len, row_num )[i] ]
                # NSTools.p( table[ri][ci], new_row, break_col( table[ri][ci], max_len, row_num ) )
            if set( new_row ) != {''}:
                new_table += [new_row]

    return new_table


def get_table_header( h_lst ):
    '''
    Parameters
    ----------
    h_lst : list<object>
    
    Returns
    -------
    string
        Table header in Tex.
    '''

    h_str = ''
    h_str += '\n' + '\\begin{center}'
    h_str += '\n' + '{\\tiny'
    h_str += '\n' + '\\begin{longtable}{|' + ( len( h_lst ) * 'c|' ) + '}'
    h_str += '\n' + '\\hline'
    h_str += '\n'
    for h in h_lst:
        h_str += str( h ) + '&'
    h_str = h_str[:-1]
    h_str += '\n' + '\\\\\\hline\\hline\\endhead'
    return h_str


def get_table_footer():
    '''
    Returns
    -------
    string
        Table footer in Tex.
    '''
    f_str = ''
    f_str += '\n' + '\\end{longtable}'
    f_str += '\n' + '}'
    f_str += '\n' + '\\end{center}'
    return f_str


def table_to_tex( h_lst, table, replace_dct = {}, col_idx = -1, max_len = 60, row_num = 10 ):
    '''
    Parameters
    ----------
    h_lst : list<string>
        A list of string defining the column 
        headers of a table.
    
    table : list<list<object>>
        A list of lists represent rows in a table.
    
    replace_dct : dict<string:string>
        A dictionary whose keys and values are strings.
    
    col_idx : int
        Index of a column.
    
    Returns
    -------
    string
        A string representing "table" in Tex format.
        The column headers are defined by "h_lst".
        The characters in each column entry are adapted using
        dictionary "replace_dct" so that key is replaced with value.  
        If the "col_idx"-th column changes at the subsequent row,
        then a horizontal line separator is added. 
        If a column entry is longer than "max_len" characters, 
        then then it is broken up in several rows, by the splitting 
        the string with respect to the comma.         
    '''
    out = ''
    out += get_table_header( h_lst )

    # split columns that are too long in several rows
    table = refine_table( table, max_len, row_num )

    # construct tex string for each row in table
    prv_val = '' if col_idx == -1 else table[0][col_idx]
    for row in table:

        # Check whether to add a horizontal separator line in table.
        # We take in account that a row may be split up in several rows.
        cur_val = '' if col_idx == -1 else row[col_idx]
        if cur_val == '':
            cur_val = prv_val
        if cur_val != prv_val:
            out += '\\hline' + '\n'
        prv_val = cur_val

        # tex code for row
        for col in row:

            # replace characters in column string
            for key in replace_dct:
                col = str( col ).replace( key, replace_dct[key] )

            # tex code for a column
            out += '$' + str( col ) + ' $' + ' &'

        # omit the last &-separator
        out = out[:-1]

        # tex code for ending a table row
        out += '\\' + '\\' + '\\hline' + '\n'

    # add footer for table in the end
    out += get_table_footer()
    return out



'''
Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
Created on Feb 7, 2017
@author: Niels Lubbes
'''
from ns_lattice.sage_interface import sage_load
from ns_lattice.sage_interface import sage_save

import inspect
import time
import sys
import os


class NSTools():
    '''
    For accessing static variables in python see for example:
    <http://stackoverflow.com/questions/68645/static-class-variables-in-python>    
    '''

    # Private dictionary object for caching result
    # used by ".get_tool_dct()" and ".save_tool_dct()".
    # If "enable_tool_dct" is false then caching in
    # disabled. This is useful for example in test
    # methods. However, it should be noted that it
    # could take a long time to compute the data.
    #
    __tool_dct = None
    __enable_tool_dct = True

    # private variable for timer
    #
    __start_time = None
    __end_time = None

    # private static variables used by ".p()"
    # If "__filter_fname_lst" equals [] then output is surpressed.
    # If "__filter_fname_lst" equals None the no output is surpressed
    #
    __filter_fname_lst = []
    __prev_filter_fname_lst = None


    @staticmethod
    def filter( filter_fname_lst ):
        '''
        It is adviced to access this method as statically as .filter().  
        See .p() for more details.
        
        Parameters
        ----------
        filter_fname_lst : list<str> 
            List of file names for Python modules.
            If None, then no output is surpressed by method ".p()". 
        '''
        NSTools.__filter_fname_lst = filter_fname_lst
        NSTools.__prev_filter_fname_lst = filter_fname_lst


    @staticmethod
    def filter_unset():
        '''
        Output via ".out" will not be surpressed.
        '''
        NSTools.__filter_fname_lst = None


    @staticmethod
    def filter_reset():
        '''
        Resets filter state to before previous ".filter_unset()" call.
        '''
        NSTools.__filter_fname_lst = NSTools.__prev_filter_fname_lst


    @staticmethod
    def p( *arg_lst ):
        '''
        Parameters
        ----------
        *arg_lst
            Variable length argument list.
        
        Returns
        -------
        string
            If ".filter_on(<fname>)" has been called and the file name
            of the calling module does not coincide with <fname>,
            then the output is surpressed and "None" is returned.
                            
            Otherwise, this method prints arguments to "sys.stdout" 
            together with reflection info from "inspect.stack()".
            Additional returns the output string.
              
            Call ".filter_off()" to turn off filter, such that
            all output is send to "sys.stdout".                                     
        '''
        # collect relevant info from stack trace
        sk_lst_lst = inspect.stack()
        file_name = os.path.basename( str( sk_lst_lst[1][1] ) )  # exclude path from file name
        line = str( sk_lst_lst[1][2] )
        method_name = str( sk_lst_lst[1][3] )

        # only output when .p() is called from module whose
        # file name is in .__filter_fname_lst
        if NSTools.__filter_fname_lst != None:
            if not file_name in NSTools.__filter_fname_lst:
                return

        # construct output string
        s = method_name + '(' + line + ')' + ': '
        for arg in arg_lst:
            s += str( arg ) + ' '

        # print output
        print( s )
        sys.stdout.flush()

        return s


    @staticmethod
    def set_enable_tool_dct( enable_tool_dct ):
        NSTools.filter_unset()
        NSTools.p( 'Caching enabled: ', enable_tool_dct )
        NSTools.filter_reset()
        NSTools.__enable_tool_dct = enable_tool_dct


    @staticmethod
    def get_tool_dct( fname = 'ns_tools' ):
        '''
        Parameters
        ----------
        fname : str
            Name of file without extension.
        
        Returns
        -------
        dct
            Sets static private variable "__tool_dct" 
            in memory from file "<local path>/<fname>.sobj"
            if called for the first time.
              
            Returns ".__tool_dct" if ".__enable_tool_dct==True" 
            and "{}" otherwise.
        '''
        if not NSTools.__enable_tool_dct:
            return {}

        path = os.path.dirname( os.path.abspath( __file__ ) ) + '/'
        file_name = path + fname
        if NSTools.__tool_dct == None:

            try:

                NSTools.p( 'Loading from:', file_name )
                NSTools.__tool_dct = sage_load( file_name )

            except Exception as e:

                NSTools.filter_unset()
                NSTools.p( 'Cannot load ".__tool_dct": ', e )
                NSTools.filter_reset()
                NSTools.__tool_dct = {}

        return NSTools.__tool_dct


    @staticmethod
    def save_tool_dct( fname = 'ns_tools' ):
        '''
        Saves ".__tool_dct" to  "fname" if ".enable_tool_dct==True" 
        otherwise do nothing.
        
        Parameters
        ----------
        fname : str
            Name of file without extension.
        '''
        if not NSTools.__enable_tool_dct:
            return

        path = os.path.dirname( os.path.abspath( __file__ ) ) + '/'
        file_name = path + fname

        NSTools.p( 'Saving to:', file_name )
        sage_save( NSTools.__tool_dct, file_name )


    @staticmethod
    def start_timer():
        '''
        Prints the current wall clock time and starts timer.
        '''
        # get time
        NSTools.__start_time = time.time()  # set static variable.

        NSTools.filter_unset()
        NSTools.p( 'start time =', NSTools.__start_time )
        NSTools.filter_reset()


    @staticmethod
    def end_timer():
        '''
        Prints wall clock time passed since last call of ".start_timer()".
        '''
        NSTools.__end_time = time.time()
        passed_time = NSTools.__end_time - NSTools.__start_time

        NSTools.filter_unset()
        NSTools.p( 'time passed =', passed_time )
        NSTools.filter_reset()




'''
Created on Aug 11, 2016
@author: Niels Lubbes
'''

from ns_lattice.sage_interface import sage_ZZ
from ns_lattice.sage_interface import sage_diagonal_matrix
from ns_lattice.sage_interface import sage_vector


class Div:
    '''Element in Neron-Severi lattice.
    
    The class represents an element in the Neron-Severi lattice
    with respect to the standard basis:
        <e0,e1,e2,...>
    

    Parameters
    ----------
    e_lst : list<sage_ZZ>
        A list of integers presents 
    
    int_mat : list<sage_matrix> 
        A matrix over ZZ of rank "len(e_lst)" represents
        the unimodular intersection product for divisor.                                  
    
    
    Attributes
    ----------            
    e_lst : list<Integer>
        List describes a divisor in terms of a standard basis. 
    
    int_mat : sage_matrix<sage_ZZ>
        Matrix defines the intersection product.
        
    '''

    # static variable
    #
    short_output = True


    # static list of intersection matrices
    #
    int_mat_lst = []


    def __init__( self, e_lst = 9 * [0], int_mat = None ):
        '''        
        Return
        ------
        Div
            Constructor (called when instantiating object).
              
            If "int_mat==None" then the default 
            diagonal matrix has signature (+-...-). 
            This matrix determines the intersection
            product of divisors.
        '''

        self.e_lst = list( e_lst )


        #
        # equal "self.int_mat" for each instantiated Div object references
        # to a unique matrix, so that no new matrix is instantiated for each
        # Div object. Maybe this is already ensured by Sage library, but just
        # to be on the safe side.
        #
        if int_mat == None:
            int_mat = sage_diagonal_matrix( sage_ZZ, [1] + ( self.rank() - 1 ) * [-1] )
        if int_mat not in Div.int_mat_lst:
            Div.int_mat_lst += [int_mat]
        idx = Div.int_mat_lst.index( int_mat )
        self.int_mat = Div.int_mat_lst[idx]



    @staticmethod
    def new( lbl, rank = 9 ):
        '''
        Parameters
        ----------
        lbl : string 
            A string with format as output of "self.get_label()".
        
        rank : int 
            Integer representing rank of Neron-Severi lattice in which "Div" lives.
        
        Returns
        -------
            The "Div" corresponding to the label 
            such that "len(self.e_lst)>=rank".
        '''

        c = Div( rank * [0] )  # zero divisor class

        if 'e' in lbl:

            s = lbl
            if 'e0' in s:

                # cases: 'e0...', '-e0...', '3e0...' or '-2e0...'
                if s[0:2] == 'e0':
                    c.e_lst = [1]
                    s = s[2:]
                elif s[0:3] == '-e0':
                    c.e_lst = [-1]
                    s = s[3:]
                else:  # '3e0...' or '-2e0...'
                    c.e_lst = [ int( s.split( 'e0' )[0] ) ]  # [4] if lbl='4h+3e...'
                    s = s.split( 'e0' )[1]  # for example '+3e2-2e5+6e7+e8'

            else:
                c.e_lst = [0]
                s = lbl

            coef_e = ''
            idx = 0
            last_i = 0  # ei
            while idx < len( s ):
                if s[idx] != 'e':
                    coef_e += s[idx]
                    idx += 1
                elif s[idx] == 'e':
                    coef_i = ''
                    idx += 1
                    while idx < len( s ) and s[idx] not in ['+', '-']:
                        coef_i += s[idx]
                        idx += 1
                    i = int( coef_i )
                    if coef_e == '-': coef_e = '-1'
                    if coef_e in ['+', '']: coef_e = '1'
                    c.e_lst += ( i - last_i - 1 ) * [0] + [int( coef_e )]
                    coef_e = ''
                    last_i = i

        else:  # label of (-2)-class

            if rank > 9:
                raise ValueError( 'For (-2)-classes we expect the rank to be at most 9: ', rank )

            # check whether the label is negative
            if lbl[0] == '-':
                neg = True
                lbl = lbl[1:]
            else:
                neg = False

            # '12' ---> e1-e2
            if len( lbl ) == 2:
                c.e_lst[ int( lbl[0] ) ] = 1
                c.e_lst[ int( lbl[1] ) ] = -1

            # '1123' ---> e0-e1-e2-e3
            elif len( lbl ) == 4 and lbl[0] == '1':
                c.e_lst[0] = int( lbl[0] )
                c.e_lst[ int( lbl[1] ) ] = -1
                c.e_lst[ int( lbl[2] ) ] = -1
                c.e_lst[ int( lbl[3] ) ] = -1

            # '212' ---> 2e0-e3-e4-...-e8
            elif len( lbl ) == 3 and lbl[0] == '2':
                c.e_lst = 9 * [-1]
                c.e_lst[0] = int( lbl[0] )
                c.e_lst[ int( lbl[1] ) ] = 0
                c.e_lst[ int( lbl[2] ) ] = 0
                if rank != 9 and set( c.e_lst[rank:] ) != set( [0] ):
                    raise ValueError( 'Rank too low for label: ', rank, lbl )
                c.e_lst = c.e_lst[:rank]

            # '308' ---> 3e0-e1-e2-...-e7-2e8
            elif len( lbl ) == 3 and lbl[0] == '3' and lbl[1] == '0':
                c.e_lst = 9 * [-1]
                c.e_lst[0] = int( lbl[0] )
                c.e_lst[ int( lbl[2] ) ] = -2

            else:  # unknown label
                raise ValueError( 'Label has incorrect format: ', lbl )

            # for example '-12'=[0,-1,1,0,0,...]
            if neg:
                c.e_lst = [ -e for e in c.e_lst ]


        # end handling label of (-2)-class


        # update rank
        c.e_lst = c.e_lst + ( rank - len( c.e_lst ) ) * [0]

        return c


    def rank( self ):
        return len( self.e_lst )


    @staticmethod
    def get_min_rank( lbl ):
        '''
        Parameters
        ----------
        lbl : string
            A string with format as output of "self.get_label()".
        
        Returns
        -------
        int
            The minimal rank of the "Div" object with a given label.
        
        Examples
        --------
        >>> get_min_rank('78')
        9
            
        >>> get_min_rank('301')
        9
            
        >>> get_min_rank('12')
        3
        '''
        d = Div.new( lbl )
        lst = [ e for e in d.e_lst ]
        while lst[-1] == 0 and lst != []:
            lst.pop()

        return len( lst )


    def get_basis_change( self, B ):
        '''
        Parameters
        ----------
        B : sage_matrix   
            A matrix whose rows correspond to generators of 
            a new basis. We assume that the intersection matrix 
            for this basis is the default diagonal matrix with 
            diagonal (1,-1,...,-1).
        
        Returns
        -------
        Div
            A new "Div" object, which represents the current divisor  
            with respect to a new basis.                
        '''
        new_int_mat = B * self.int_mat * B.T
        new_e_lst = self.mat_mul( ~( B.T ) ).e_lst

        return Div( new_e_lst, new_int_mat )


    def __get_minus_two_label( self ):
        '''
        Private helper method for "get_label()"
        
        Parameters
        ----------
            self : Div 
                self*self==-2 and self.rank<=9.
        
        Returns
        -------
        string 
            See output documents for self.get_label()
        '''

        if self * self != -2 or self.rank() > 9:
            raise ValueError( 'Unexpected input for __get_mt_label: ', self.e_lst )

        # first non-zero coefficient negative?
        neg = [e < 0 for e in self.e_lst if e != 0][0]

        # check whether the label should start with minus symbol
        if neg:
            tmp = [-e for e in self.e_lst]
        else:
            tmp = self.e_lst

        # set of non-zero coefficients for ei.
        oset = set( [ e for e in tmp[1:] if e != 0 ] )

        # e1-e2 ---> '12'
        if tmp[0] == 0 and oset == set( [1, -1] ):
            lbl = ''
            for i in range( 1, len( tmp ) ):
                if tmp[i] != 0:
                    lbl += str( i )

        # e0-e1-e2-e3 ---> '1123'
        elif tmp[0] == 1 and oset == set( 3 * [-1] ):
            lbl = '1'
            for i in range( 1, len( tmp ) ):
                if tmp[i] != 0:
                    lbl += str( i )

        # 2e0-e3-e4-...-e8 ---> '212'
        elif tmp[0] == 2 and oset == set( 6 * [-1 ] ):
            lbl = '2'
            for i in range( 1, len( tmp ) ):
                if tmp[i] == 0:
                    lbl += str( i )

        # 3e0-e1-e2-...-e7-2e8 ---> '308'
        elif tmp[0] == 3 and oset == set( 7 * [-1 ] + [-2] ):
            lbl = '30'
            for i in range( 1, len( tmp ) ):
                if tmp[i] == -2:
                    lbl += str( i )

        if neg:
            lbl = '-' + lbl  # for example: 12 --> -12

        return lbl


    def get_abbr_label( self ):
        '''
        Returns
        -------
        string        
            We describe the output label in terms of  examples.
            
                > e1                --->  'e1'
                > e1-e2             --->  'e12'                  
                > 2e0-e1-e2-e4-e5   --->  '2e1245'
                > e0-e1             --->  '1e1'  
        
            This options only works for special cases.
            The cases which are covered are (-1)- and (-2)-classes, 
            and classes of conical families on weak Del Pezzo surfaces,
            with respect to the basis with intersection product
            defined by the diagonal matrix with diagonal (1,-1,...,-1).                        
        '''
        np1 = len( [e for e in self.e_lst[1:] if e == 1] )
        nm1 = len( [e for e in self.e_lst[1:] if e == -1] )
        n01 = len( [e for e in self.e_lst[1:] if e > 1 or e < -1] )

        if n01 == 0 and self[0] in range( 0, 10 ):

            # e1
            if self[0] == 0 and np1 == 1 and nm1 == 0:
                return 'e' + str( self.e_lst.index( 1 ) )

            # e1-e2
            if self[0] == 0 and np1 == 1 and nm1 == 1:
                return 'e' + str( self.e_lst.index( 1 ) ) + str( self.e_lst.index( -1 ) )

            # 2h-e1-e2-e3-e4-e5 or h-e1
            if self[0] in range( 0, 10 ) and np1 == 0 and nm1 > 0:
                lbl = str( self[0] ) + 'e'
                for i in range( 1, len( self.e_lst ) ):
                    if self[i] != 0:
                        lbl += str( i )
                return lbl

        raise ValueError( 'Input is not treated by this function (use get_label() instead):', self.e_lst )


    def get_label( self, abbr = False ):
        '''
        Parameters
        ----------
            abbr : boolean 
            
        Returns
        -------
        string
            We describe the output label in terms of examples.
                
            If "abbr==True" and self*self==-2 and self.rank()<=9:
            
                >  e1-e2                ---> '12'
                > -e1+e2                ---> '-12'
                >  e0-e1-e2-e3          ---> '1123'
                >  2e0-e3-e4-...-e8     ---> '212'
                >  3e0-e1-e2-...-e7-2e8 ---> '308'
                > -3e0+e1+e2+...+e7+2e8 ---> '-308'  
                
            For the remaining cases not treated above:
            
                > 3e0-2e1-13e2-4e3  ---> '3e0-2e1-13e2-4e3'                                                    
        '''

        divK = Div( [-3] + ( self.rank() - 1 ) * [1] )

        # treat cases for (-2)-label
        #
        if abbr and self * self == -2 and self.rank() <= 9 and self * divK == 0:
            return self.__get_minus_two_label()


        # from this point on we treat the general case
        #
        lbl = ''
        for i in range( 0, len( self.e_lst ) ):
            val = self[i]
            if val != 0:

                if val == 1:
                    if lbl != '': lbl += '+'
                elif val == -1:
                    lbl += '-'
                else:
                    if val > 1 and lbl != '':
                        lbl += '+'
                    lbl += str( val )

                lbl += 'e' + str( i )

        return lbl


    def mat_mul( self, M ):
        '''
        Parameters
        ----------
            M : sage_matrix
                A matrix with self.rank() columns.
        
        Returns
        -------
        Div
            Returns a "Div" object that is a result of 
            applying the linear transformation corresponding
            to "M" to itself. 
        '''
        v = sage_vector( self.e_lst ).column()
        return Div( ( M * v ).list() )


    def int_mul( self, n ):
        '''
        Parameters
        ----------
            n : int
        
        Returns
        -------
        Div
            Returns a "Div" object that is a result of multiplying 
            with the scalar "n". 
        '''
        return self.mat_mul( sage_diagonal_matrix( self.rank() * [n] ) )


    # operator overloading for ==
    def __eq__( self, other ):
        return self.e_lst == other.e_lst


    # operator overloading for !=
    def __ne__( self, other ):
        return not self.__eq__( other )


    # operator overloading for <
    # Used for sorting lists of "Div"-objects:
    #     <http://stackoverflow.com/questions/1227121/compare-object-instances-for-equality-by-their-attributes-in-python>
    def __lt__( self, other ):
        '''
        Parameters
        ----------
            other : Div

        Returns
        -------
        bool
 
            Here are some examples to explain 
            the ordering we use for div classes
                            
            e1 < e2
            e0-e1-e2 < e0-e1-e3
            1123 < 308  
            1123 < 1124
            12 < 1123 
            12 < 13
            12 < 34
        '''
        if self.rank() != other.rank():
           return self.rank() < other.rank()

        a = self.e_lst
        b = other.e_lst

        if sum( a ) == sum( b ) == 1 and set( a ) == set( b ) == {0, 1}:
            return b < a  # e1 < e2

        a = [a[0]] + [ -elt for elt in reversed( a[1:] )]
        b = [b[0]] + [ -elt for elt in reversed( b[1:] )]

        return a < b  # lexicographic order


    # operator overloading for *
    def __mul__( self, div ):
        '''
        Parameters
        ----------
        div : Div
        
        Returns
        -------
        Div
            The intersection product of "self" and 
            "div" wrt. to matrix "self.int_mat".
        '''

        row_vec = sage_vector( sage_ZZ, self.e_lst ).row()
        col_vec = sage_vector( sage_ZZ, div.e_lst ).column()
        mat = self.int_mat

        v = row_vec * mat * col_vec

        return v[0][0]


    # operator overload for +
    def __add__( self, div ):
        v = sage_vector( sage_ZZ, self.e_lst ) + sage_vector( sage_ZZ, div.e_lst )
        return Div( list( v ) )


    # operator overload for -
    def __sub__( self, div ):
        v = sage_vector( sage_ZZ, self.e_lst ) - sage_vector( sage_ZZ, div.e_lst )
        return Div( list( v ) )


    # operator overloading for []
    def __getitem__( self, index ):
        return self.e_lst[index]


    # operator overloading for []
    def __setitem__( self, index, item ):
        self.e_lst[index] = item


    # overloading for str(.): human readable string representation of object
    def __str__( self ):
        if Div.short_output:
            return self.get_label()
        else:
            return str( self.e_lst )


    # overloading "__repr__()" as well, since python call this for Div objects in a list
    def __repr__( self ):
        return self.__str__()


    # so that lists of this object can be used with set()
    def __hash__( self ):
        return hash( self.__str__() + '__' + str( self.rank() ) )



'''
Use of this source code is governed by a 
MIT-style license that can be found in the 
LICENSE file.
Created on Jul 12, 2017
@author: Niels Lubbes

Sage has a complicated import structure and it 
is not possible to simply import each need 
module. It seems that "from sage.all import *" 
is the only option. Therefore we introduce an 
interface to Sage so that in the code, it is 
clear, which libraries of Sage we use. Moreover, 
we specify below from which modules in the Sage 
library we import. 

We explain the naming scheme with the following 
two examples. The interface method for 
"PolynomialRing()" is called 
"sage_PolynomialRing()". However the interface 
method for "sage_eval()" is not called 
"sage_sage_eval()" but instead "sage__eval()". 
The variable "ZZ" is called "sage_ZZ".


For the Parameters section in the documentation 
of types we will use the following abbrevations:

sage_POLY: 
    sage.rings.polynomial.multi_polynomial_element.MPolynomial_polydict
    The type of an element in sage_PolynomialRing 

sage_RING:
    sage.rings.*
    The type of a ring. For example sage_QQ or sage_ZZ or sage_NumberField.

sage_GRAPH:
    sage.graphs.graph
    The type of a Graph.
'''

from sage.all import *
from sage.structure.sage_object import register_unpickle_override

#################################################
# sage.structure                                #
#################################################

# from sage.structure.proof.proof import proof
sage_proof = proof

# from sage.structure.sage_object import save
def sage_save( *args, **kwargs ):
    return save( *args, **kwargs )

# from sage.structure.sage_object import load
def sage_load( *args, **kwargs ):
    return load( *args, **kwargs )

# from sage.structure.sage_object import register_unpickle_override
def sage_register_unpickle_override( *args, **kwargs ):
    register_unpickle_override( *args, **kwargs )


#################################################
# sage.misc                                     #
#################################################

# from sage.misc.sage_eval import sage_eval
def sage__eval( *args, **kwargs ):
    return sage_eval( *args, **kwargs )

# from sage.misc.functional import n
def sage_n( *args, **kwargs ):
    return n( *args, **kwargs )


#################################################
# sage.symbolic                                 #
#################################################

# from sage.symbolic.ring import SR
sage_SR = SR

# from sage.symbolic.relation import solve
def sage_solve( *args, **kwargs ):
    return solve( *args, **kwargs )


#################################################
# sage.rings                                    #
#################################################

# from sage.rings.integer_ring import ZZ
sage_ZZ = ZZ

# from sage.rings.rational_field import QQ
sage_QQ = QQ

# import sage.rings.invariant_theory
sage_invariant_theory = invariant_theory

# from sage.rings.fraction_field import FractionField
def sage_FractionField( *args, **kwargs ):
    return FractionField( *args, **kwargs )

# from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
# http://doc.sagemath.org/html/en/reference/polynomial_rings/sage/rings/polynomial/polynomial_ring_constructor.html
def sage_PolynomialRing( *args, **kwargs ):
    return PolynomialRing( *args, **kwargs )

# from sage.rings.number_field.number_field import NumberField
def sage_NumberField( *args, **kwargs ):
    return NumberField( *args, **kwargs )


#################################################
# sage.modules                             #
#################################################

# from sage.modules import VectorSpace
def sage_VectorSpace( *args, **kwargs ):
    return VectorSpace( *args, **kwargs )


#################################################
# sage.matrix                                   #
#################################################

# from sage.matrix.constructor import matrix
def sage_matrix( *args, **kwargs ):
    return matrix( *args, **kwargs )

# from sage.matrix.constructor import identity_matrix
def sage_identity_matrix( *args, **kwargs ):
    return identity_matrix( *args, **kwargs )

# from sage.matrix.constructor import diagonal_matrix
def sage_diagonal_matrix( *args, **kwargs ):
    return diagonal_matrix( *args, **kwargs )

# from sage.matrix.constructor import vector
def sage_vector( *args, **kwargs ):
    return vector( *args, **kwargs )


#################################################
# sage.arith                                    #
#################################################

# from sage.arith.misc import factor
def sage_factor( *args, **kwargs ):
    return factor( *args, **kwargs )

# from sage.arith.misc import gcd
def sage_gcd( *args, **kwargs ):
    return gcd( *args, **kwargs )


#################################################
# sage.calculus                                 #
#################################################

# from sage.calculus.functional import diff
def sage_diff( *args, **kwargs ):
    return diff( *args, **kwargs )

# from sage.calculus.functional import expand
def sage_expand( *args, **kwargs ):
    return expand( *args, **kwargs )

# from sage.calculus.var import var
def sage_var( *args, **kwargs ):
    return var( *args, **kwargs )


#################################################
# sage.combinat                                 #
#################################################

# from sage.combinat.composition import Compositions
def sage_Compositions( *args, **kwargs ):
    return Compositions( *args, **kwargs )

# from sage.combinat.combination import Combinations
def sage_Combinations( *args, **kwargs ):
    return Combinations( *args, **kwargs )

# from sage.combinat.partitions import Partitions
def sage_Partitions( *args, **kwargs ):
    return Partitions( *args, **kwargs )

# from sage.combinat.permutations import Partitions
def sage_Permutations( *args, **kwargs ):
    return Permutations( *args, **kwargs )

# from sage.subset import Subsets
def sage_Subsets( *args, **kwargs ):
    return Subsets( *args, **kwargs )

# from sage.combinat.root_system.root_system import RootSystem
def sage_RootSystem( *args, **kwargs ):
    return RootSystem( *args, **kwargs )


#################################################
# sage.graphs                                   #
#################################################

# from sage.graphs.graph import Graph
def sage_Graph( *args, **kwargs ):
    return Graph( *args, **kwargs )









'''
Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
Created on Feb 9, 2017
@author: Niels Lubbes
'''
from ns_lattice.sage_interface import sage_identity_matrix
from ns_lattice.sage_interface import sage_matrix
from ns_lattice.sage_interface import sage_ZZ
from ns_lattice.sage_interface import sage_Permutations
from ns_lattice.sage_interface import sage_Subsets

from ns_lattice.class_div import Div

from ns_lattice.div_in_lattice import get_indecomp_divs
from ns_lattice.div_in_lattice import get_ak
from ns_lattice.div_in_lattice import get_divs

from ns_lattice.class_dp_lattice import DPLattice

from ns_lattice.class_eta import ETA

from ns_lattice.class_ns_tools import NSTools


def get_bases_lst( a_lst, M, d_lst, m1_lst, perm = False ):
    '''
    Returns a list of basis with specified generators.
    
    Parameters
    ----------
    a_lst : list<Div>
        A list of linear independent Div objects of 
        the same rank with 3<=rank<=9.
        It is required that
        "set(a_lst)==set([ a.mat_mul(M) for a in a_lst ])".
    
    M : sage_matrix<sage_ZZ>
        A unimodular matrix representing an involution.
    
    d_lst : list<Div>
        A list of Div objects d of the same rank as any
        element in "a_lst", so that "d*k==0" and "d*d==-2".
        These represent a root basis for the indecomposable 
        (-2)-classes in the Neron-Severi lattice of a 
        weak del Pezzo surface.   

    m1_lst : list<Div>
        A list of Div objects d of the same rank as any
        element in "a_lst", so that "d*k==d*d==-1".
        These represent (-1)-classes in the Neron-Severi 
        lattice of a weak del Pezzo surface.     
    
    perm : bool
        If False, then we consider two bases the same if the 
        generators of the first basis can be obtained from 
        the second basis via a permutation matrix.
    
    Returns
    -------
    list<tuple<Div>>
        A list of tuples of Div objects. Each tuple of Div objects
        represents a basis for the Neron-Severi lattice determined
        by d_lst and m1_lst. The bases are of the form          
            < a1,...,as, b1,...,bt >
        with the following property
            * a1,...,as are defined by the input "a_lst"
            * bi is an element in m1_lst such that bi*bj=am*bi=0 
              for all 1<=i<j<=t and 1<=m<=s              
        If "a_lst==[]" then "[[]]" is returned.   
               
    '''
    key = 'get_bases_lst__' + str( ( a_lst, M, d_lst, m1_lst, perm ) ) + '__' + str( M.rank() )
    if key in NSTools.get_tool_dct():
        return NSTools.get_tool_dct()[key]


    if a_lst == []:
        return [[]]

    if len( a_lst ) == a_lst[0].rank():
        return [tuple( a_lst )]

    e_lst = []
    for m1 in get_indecomp_divs( m1_lst, d_lst ):
        if set( [ m1 * a for a in a_lst ] ) != {0}:
            continue
        if m1 * m1.mat_mul( M ) > 0:
            continue
        e_lst += [m1]

    bas_lst = []
    for e in e_lst:

        Me = e.mat_mul( M )
        new_d_lst = [ d for d in d_lst if d * e == d * Me == 0 ]
        new_m1_lst = [ m1 for m1 in m1_lst if m1 * e == m1 * Me == 0 ]
        add_lst = [e]
        if e != Me: add_lst += [Me]
        bas2_lst = get_bases_lst( a_lst + add_lst, M, new_d_lst, new_m1_lst, perm )

        if perm:
            bas_lst += bas2_lst
        else:
            for bas2 in bas2_lst:
                found = False
                for bas in bas_lst:
                    # check whether the two bases are the same up to
                    # permutation of generators
                    if set( bas ) == set( bas2 ):
                        found = True
                        break  # break out of nearest for loop
                if not found:
                    NSTools.p( 'found new basis: ', bas2, ', bas2_lst =', bas2_lst )
                    bas_lst += [bas2]

    # cache output
    NSTools.get_tool_dct()[key] = bas_lst
    NSTools.save_tool_dct()

    return bas_lst


def get_webs( dpl ):
    '''
    Returns lists of families of conics for each possible complex basis change.
    The n-th family in each list correspond to a fixed family wrt.
    different bases for each n. 
    
    Parameters
    ----------
    dpl : DPLattice
        Represents the Neron-Severi lattice of a weak del Pezzo surface. 
        
    Returns
    -------
    list<list<Div>>
        A list of lists of Div objects. 
        Each Div object f has the property that 
        f*(3e0-e1-...-er)=2, f*f==0 and f*d>=0 for all d in dpl.d_lst.
        Such a Div object corresponds geometrically to a family of conics.
        For each index i, the i-th entry of each list of Div object corresponds
        to the same family of conics.          
    '''
    key = 'get_webs__' + str( dpl ).replace( '\n', '---' )
    if key in NSTools.get_tool_dct():
        return NSTools.get_tool_dct()[key]

    ak = get_ak( dpl.get_rank() )
    all_m1_lst = get_divs( ak, 1, -1, True )
    akc, cc = ( 3, 1 )
    M = sage_identity_matrix( dpl.get_rank() )

    fam_lst_lst = []
    for e0 in get_divs( ak, akc, cc, True ):
        NSTools.p( 'e0 =', e0 )
        for B_lst in get_bases_lst( [e0], M, dpl.d_lst, all_m1_lst, True ):
            B = sage_matrix( sage_ZZ, [ d.e_lst for d in B_lst ] )
            dplB = dpl.get_basis_change( B )
            fam_lst_lst += [ dplB.real_fam_lst ]

    # reduce fam_lst
    pat_lst_lst = []
    rfam_lst_lst = []
    for fam_lst in fam_lst_lst:
        pat_lst = [ 0  if fam[0] != 1 else 1 for fam in fam_lst ]
        if pat_lst not in pat_lst_lst:
            pat_lst_lst += [ pat_lst ]
            rfam_lst_lst += [ fam_lst ]


    # cache output
    NSTools.get_tool_dct()[key] = rfam_lst_lst
    NSTools.save_tool_dct()

    return rfam_lst_lst


def contains_perm( f_lst_lst, c_lst ):
    '''
    Parameters
    ----------
    f_lst_lst : list<list<Div>>
        A list of lists containing Div objects.
    
    c_lst : list<Div>
        A list of Div objects
    
    Returns:
    --------
    bool
        Returns True if after a permutation of the generators
        (e1,...,er) the list c_lst is contained in f_lst_lst. 
        For example if c_lst equals [ e0-e1, 2e0-e2-e3-e4-e5 ] 
        then is contained in [ ..., [e0-e2, 2e0-e1-e3-e4-e5], ... ].
    
    '''
    if c_lst == []:
        return [] in f_lst_lst

    for perm in sage_Permutations( range( c_lst[0].rank() - 1 ) ):
        pc_lst = [ Div( [c[0]] + [ c[i + 1] for i in perm ], c.rank() ) for c in c_lst ]
        for f_lst in f_lst_lst:
            if set( f_lst ) == set( pc_lst ):
                return True

    return False


def triples( dpl, mval ):
    '''
    Parameters
    ----------
    dpl  : DPLattice
    mval : integer
    
    Returns
    -------
    list<(Div,Div,Div)>
        List of triples in "dpl.fam_lst":
            [ (a,b,c),... ]
        so that 
            (1) There does not exists e in "dpl.m1_lst"
                with the property that a*e==b*e==c*e==0.
            (2) 1 <= max( a*b, a*c, b*c ) <= mval.
    '''
    key = 'triples__' + str( dpl ).replace( '\n', '---' ) + '---' + str( mval )
    if key in NSTools.get_tool_dct():
        return NSTools.get_tool_dct()[key]

    f_lst = dpl.fam_lst
    e_lst = dpl.m1_lst

    # obtain list of triples (a,b,c) in f_lst
    # that are not orthogonal to any element in e_lst
    t_lst = []
    idx_lst_lst = sage_Subsets( range( len( f_lst ) ), 3 )
    eta = ETA( len( idx_lst_lst ), 500000 )
    for idx_lst in idx_lst_lst:
        eta.update( 't_lst' )

        t = [ f_lst[idx] for idx in idx_lst ]
        if t[0] * t[1] > mval: continue
        if t[0] * t[2] > mval: continue
        if t[1] * t[2] > mval: continue

        # elements in f_lst correspond to divisor classes of curves on a
        # surface and thus t[i]*t[j]>=1 for all i,j \in {0,1,2} so that i!=j.

        cont = False
        for e in e_lst:
            if [f * e for f in t] == [0, 0, 0]:
                cont = True
                break
        if cont: continue

        if not contains_perm( t_lst, t ):
            t_lst += [t]

    NSTools.p( 't_lst =', t_lst )

    # cache output
    NSTools.get_tool_dct()[key] = t_lst
    NSTools.save_tool_dct()

    return t_lst



'''
Created on Aug 15, 2016
@author: Niels Lubbes

This module is for classifying real structures and singularities 
of weak Del Pezzo surfaces of degree between 1 and 7.

'''
import time

from ns_lattice.sage_interface import sage_identity_matrix
from ns_lattice.sage_interface import sage_ZZ
from ns_lattice.sage_interface import sage_QQ
from ns_lattice.sage_interface import sage_Subsets
from ns_lattice.sage_interface import sage_VectorSpace
from ns_lattice.sage_interface import sage_vector

from ns_lattice.div_in_lattice import get_divs
from ns_lattice.div_in_lattice import get_indecomp_divs
from ns_lattice.div_in_lattice import get_ak

from ns_lattice.dp_root_bases import get_ext_graph
from ns_lattice.dp_root_bases import get_dynkin_type
from ns_lattice.dp_root_bases import get_root_bases_orbit
from ns_lattice.dp_root_bases import is_root_basis

from ns_lattice.dp_involutions import basis_to_involution
from ns_lattice.dp_involutions import is_integral_involution

from ns_lattice.class_ns_tools import NSTools

from ns_lattice.class_div import Div


class DPLattice:
    '''
    Represents an equivalence class of the Neron-Severi lattice 
    of a real weak del Pezzo surface, together with an involution "M"
    and a set of effective (-2)-classes "d_lst". The effective (-2)-classes
    form the basis of a root system.
    
        ( ZZ<e0,e1,...,er>, M, d_lst )
    
    From these objects it is possible to compute the remaining attributes of 
    this class. 
    
    If <e0,e1,...,er> is a basis for the Neron-Severi lattice of the 
    projective plane P^2 blown up in r points then the the canonical 
    class k equals 
        k=-3e0+e1+...+er.
    The intersection product is in this case -e0^2=e1^2=...=er^2=-1 with
    remaining intersections zero.
    
    Otherwise if <e0,e1,...,er> is a basis for the Neron-Severi lattice of the 
    P^1xP^1 blown up in r points then the the canonical 
    class k equals 
        k=-2*(e0+e1).
    The intersection product is in this case -h*e1=e2^2=...=er^2=-1 with
    remaining intersections zero.         
    

    Attributes
    ----------
    M : sage_matrix<sage_ZZ>
        A matrix which correspond to an involution of the lattice
        <e0,e1,...,er> with r=rank-1 and 2 <= r <= 8.
    
    Md_lst : list<Div>
        A list of "Div" objects that correspond to the eigenvectors
        of eigenvalue 1 of M. These "Div" objects form a basis of
        a root subsystem.

    Mtype : string
        A String that denotes the  Dynkin type of "Md_lst".
        
    d_lst : list<Div>
        A list of "Div" objects d such that d*d==-2 and d*k=0
        where k denotes the canonical class. These elements 
        represent effective (-2)-classes.

    type : string
        A String that denotes the Dynkin type of "d_lst".
 
    m1_lst : list<Div>
        A list of "Div" objects "m" such that
        m*m==-1==m*k and m*d>=0 for all d in d_lst, 
        where k denotes the canonical class.        
        These elements represent (-1)-classes that cannot be written 
        as the sum of two effective classes. 
        In other words, the classes are indecomposable.
        
    fam_lst : list<Div>
        A list of "Div" objects "f" such that 
        f*f==0, f*(-k)==2 and m*d>=0 
        for all d in d_lst, where k denotes the canonical class.
                
    real_d_lst : list<Div>
        A list "Div" objects that represent indecomposable and 
        real (-2)-classes. Thus these classes are send to itself by M.
        Geometrically these classes correspond to real isolated singularities.
        
        
    real_m1_lst : list<Div>
        A list "Div" objects that represent indecomposable and 
        real (-1)-classes. Thus these classes are send to itself by M.
        Geometrically these classes correspond to real lines.
        
        
    self.real_fam_lst : list<Div>
        A list "Div" objects that represent real classes in "self.fam_lst".
        Thus these classes are send to itself by M.
        Geometrically these classes correspond to a real families of conics        
   
    self.or_lst : list<Div>
        A list of "Div" objects that represents roots that are orthogonal to
        "self.d_lst".  

    self.sr_lst : list<Div>
        A list of "Div" objects that represents roots that are contained in
        the subspace spanned by "self.d_lst".  
    
    self.G : sage_Graph
        The Cremona invariant for the current lattice.
    '''

    # Static variable. If False, then equivalent real structures with
    # different Dynkin types are considered equivalent. Equivalent real
    # structures can have several Dynkin types. We set this variable to
    # False in order to register this. See also get_reduced_cls()
    weak_equality = False

    def __init__( self, d_lst, Md_lst, M ):
        '''
        Constructor.
        
        Returns
        -------
        DPLattice
            A DPLattice class whose attributes are set according to input:
                * DPLattice.M
                * DPLattice.Md_lst
                * DPLattice.d_lst
            The remaining attributes of DPLattice can be computed 
            from these attributes.
                    
            In order for this object to make sense, it is required that the 
            involution "M" preserves "d_lst" as a set. Geometrically this 
            means that the involution sends isolated singularities to isolated 
            singularities.  
        '''

        self.d_lst = d_lst
        self.Md_lst = Md_lst
        self.M = M

        self.m1_lst = None
        self.fam_lst = None
        self.real_d_lst = None
        self.real_m1_lst = None
        self.real_fam_lst = None
        self.Mtype = None
        self.type = None
        self.or_lst = None
        self.sr_lst = None
        self.G = None


    def set_attributes( self, level = 9 ):
        '''
        Sets attributes of this object, depending
        on the input level.
        For constructing a classification we instantiate
        many DPLattice objects. This method allows us 
        to minimize the number of attributes that computed
        (thus we use lazy evaluation).
        
        Parameter
        ---------
        self: DPLattice
            At least self.M, self.Md_lst and self.d_lst        
            should be initialized.
        
        level : int
            A positive number.
        '''

        # M, Md_lst and d_lst are set.

        if self.m1_lst == None:
            all_m1_lst = get_divs( get_ak( self.get_rank() ), 1, -1, True )
            self.m1_lst = get_indecomp_divs( all_m1_lst, self.d_lst )

        if level < 1: return

        if self.fam_lst == None:
            all_fam_lst = get_divs( get_ak( self.get_rank() ), 2, 0, True )
            self.fam_lst = get_indecomp_divs( all_fam_lst, self.d_lst )

        if level < 2: return

        if self.real_d_lst == None:
            self.real_d_lst = [ d for d in self.d_lst if d.mat_mul( self.M ) == d ]

        if level < 3: return

        if self.real_m1_lst == None:
            self.real_m1_lst = [ m1 for m1 in self.m1_lst if m1.mat_mul( self.M ) == m1 ]

        if level < 4: return

        if self.real_fam_lst == None:
            self.real_fam_lst = [ f for f in self.fam_lst if f.mat_mul( self.M ) == f ]

        if level < 5: return

        if self.or_lst == None:
            self.or_lst = []
            for m2 in get_divs( get_ak( self.get_rank() ), 0, -2, True ):
                if [m2 * d for d in self.d_lst] == len( self.d_lst ) * [0]:
                    self.or_lst += [m2]

        if level < 6: return

        if self.sr_lst == None:
            V = sage_VectorSpace( sage_QQ, self.get_rank() )
            W = V.subspace( [d.e_lst for d in self.d_lst] )
            self.sr_lst = []
            for m2 in get_divs( get_ak( self.get_rank() ), 0, -2, True ):
                if sage_vector( m2.e_lst ) in W:
                    self.sr_lst += [ m2 ]

        if level < 7: return

        if self.type == None:
            self.type = get_dynkin_type( self.d_lst )

        if level < 8: return

        if self.Mtype == None:
            self.Mtype = get_dynkin_type( self.Md_lst )

        if level < 9: return

        if self.G == None:
            self.G = get_ext_graph( self.d_lst + self.m1_lst, self.M )




    def get_rank( self ):
        '''
        Parameters
        ----------
        self : DPLattice
            We expect self.M != None.
        
        Returns
        -------
        int
            Integer denoting rank of lattice.
        '''
        return self.M.dimensions()[0]


    def get_degree( self ):
        '''
        Parameters
        ----------
        self : DPLattice
            We expect self.M != None.        
        
        Returns
        -------
        int
            Integer denoting the degree of weak del Pezzo surface with
            "self" its corresponding Neron-Severi lattice.
        '''
        return 10 - self.get_rank()


    def get_numbers( self ):
        '''
        Parameters
        ----------
        self : DPLattice
        
        Returns
        -------
        list<int>
            List of 6 integers:
                
                0: #indecomposable (-2)-classes 
                1: #indecomposable (-1)-classes
                2: #families of conics
                
                3: #real effective (-2)-classes 
                4: #real indecomposable (-1)-classes
                5: #real families of conics                
            
            where # stands for number of.
            
            Note that a divisor class is indecomposable 
            if it is effective and cannot be written as 
            the sum of two effective classes.
        '''
        self.set_attributes( 6 )
        return ( len( self.d_lst ),
                 len( self.m1_lst ),
                 len( self.fam_lst ),
                 len( self.real_d_lst ),
                 len( self.real_m1_lst ),
                 len( self.real_fam_lst ) )


    def contains_fam_pair( self ):
        '''
        Parameters
        ----------
        self : DPLattice
        
        Returns
        -------
        bool
            True if self.real_fam_lst contains two Div classes 
            with intersection one. Geometrically this means that a 
            weak del Pezzo surface with a Neron-Severi lattice that
            is isomorphic to this one, must be birational to P1xP1
            (ie. fiber product of the projective line with itself).             
        '''
        self.set_attributes( 6 )
        for f1 in self.real_fam_lst:
            for f2 in self.real_fam_lst:
                if f1 * f2 == 1:
                    return True
        return False


    def get_marked_Mtype( self ):
        '''
        We mark Mtype with a '-symbol to distinguish between real 
        structures of the same Dynkin type that are not conjugate.
        '''
        if self.get_degree() not in [6, 4, 2]:
            return self.Mtype

        self.set_attributes( 8 )
        if ( self.get_degree(), self.Mtype ) not in [ ( 6, 'A1' ), ( 4, '2A1' ), ( 2, '3A1' ) ]:
            return self.Mtype

        mark = ''
        if list( self.M.T[0] ) != [1] + ( self.get_rank() - 1 ) * [0]:
            # in this case e0 is not send to e0 by the involution self.M
            mark = "'"

        return self.Mtype + mark


    def get_basis_change( self, B ):
        '''
        Parameters
        ----------
        self : DPLattice
            
        B : sage_matrix<sage_ZZ>   
            A matrix whose rows correspond to generators of 
            a new basis. We assume that the intersection
            matrix for this basis is the default
            diagonal matrix with diagonal (1,-1,...,-1).
        
        Returns
        -------
        DPLattice
            A new "DPLattice" object, which represents the current  
            lattice with respect to a new basis.
                
        '''
        self.set_attributes( 6 )

        d_lst_B = [ d.get_basis_change( B ) for d in self.d_lst ]
        Md_lst_B = [ Md.get_basis_change( B ) for Md in self.Md_lst ]
        M_B = ~( B.T ) * self.M * ( B.T )  # ~B is inverse of B, new involution after coordinate change

        dpl = DPLattice( d_lst_B, Md_lst_B, M_B )
        dpl.Mtype = self.Mtype
        dpl.type = self.type
        dpl.m1_lst = [ m1.get_basis_change( B ) for m1 in self.m1_lst ]
        dpl.fam_lst = [ fam.get_basis_change( B ) for fam in self.fam_lst ]
        dpl.real_d_lst = [ d.get_basis_change( B ) for d in self.real_d_lst ]
        dpl.real_m1_lst = [ m1.get_basis_change( B ) for m1 in self.real_m1_lst ]
        dpl.real_fam_lst = [ fam.get_basis_change( B ) for fam in self.real_fam_lst ]

        return dpl


    @staticmethod
    def get_cls_root_bases( rank = 9 ):
        '''
        See [Algorithm 5, http://arxiv.org/abs/1302.6678] for more info. 
        
        Parameters
        ----------
        rank : int
            An integer in [3,...,9].    
        
        Returns
        -------
        list<DPLattice>
            A list of "DPLattice" objects dpl such that dpl.d_lst 
            is the bases of a root subsystem and dpl.Mtype == A0. 
            The list contains exactly one representative for all 
            root subsystems up to equivalence.  
             
            The list represents a classification of root 
            subsystems of the root system with Dynkin type either:
                A1, A1+A2, A4, D5, E6, E7 or E8,
            corresponding to ranks 3, 4, 5, 6, 7, 8 and 9 respectively 
            (eg. A1+A2 if rank equals 4, and E8 if rank equals 9).
            Note that the root systems live in a subspace of 
            the vector space associated to the Neron-Severi lattice 
            of a weak Del Pezzo surface.
        '''
        # classification of root bases in cache?
        key = 'get_cls_root_bases_' + str( rank )
        if key in NSTools.get_tool_dct():
            return NSTools.get_tool_dct()[key]

        A = [ 12, 23, 34, 45, 56, 67, 78]
        B = [ 1123, 1145, 1456, 1567, 1678, 278 ]
        C = [ 1127, 1347, 1567, 234, 278, 308 ]
        D = [ 1123, 1345, 1156, 1258, 1367, 1247, 1468, 1178 ]

        dpl_lst = []
        for ( lst1, lst2 ) in [ ( A, [] ), ( A, B ), ( A, C ), ( [], D ) ]:

            # restrict to divisors in list, that are of rank at most "max_rank"
            lst1 = [ Div.new( str( e ), rank ) for e in lst1 if rank >= Div.get_min_rank( str( e ) ) ]
            lst2 = [ Div.new( str( e ), rank ) for e in lst2 if rank >= Div.get_min_rank( str( e ) ) ]

            # data for ETA computation
            total = len( sage_Subsets( range( len( lst1 ) ) ) )
            total *= len( sage_Subsets( range( len( lst2 ) ) ) )
            counter = 0
            ival = 20

            NSTools.p( 'inside loop: ', lst1, lst2, ', rank = ', rank )

            # loop through the lists
            for idx2_lst in sage_Subsets( range( len( lst2 ) ) ):
                for idx1_lst in sage_Subsets( range( len( lst1 ) ) ):

                    # ETA
                    if counter % ival == 0:
                        start = time.time()
                    counter += 1
                    if counter % ival == 0:
                        passed_time = time.time() - start
                        NSTools.p( 'ETA in minutes =', passed_time * ( total - counter ) / ( ival * 60 ), idx1_lst, idx2_lst, ', rank =', rank )

                    # construct d_lst as a union
                    d_lst = []
                    d_lst += [ lst1[idx1] for idx1 in idx1_lst ]
                    d_lst += [ lst2[idx2] for idx2 in idx2_lst ]

                    if not is_root_basis( d_lst ):
                        continue

                    Md_lst = []
                    M = sage_identity_matrix( sage_QQ, rank )
                    dpl = DPLattice( d_lst, Md_lst, M )

                    if dpl not in dpl_lst:
                        dpl.set_attributes()
                        dpl_lst += [dpl]

        # cache output
        dpl_lst.sort()
        NSTools.get_tool_dct()[key] = dpl_lst
        NSTools.save_tool_dct()

        return dpl_lst


    @staticmethod
    def get_cls_involutions( rank = 9 ):
        '''
        Outputs a list representing a classification of root 
        subsystems that define unimodular involutions on the 
        Neron-Severi lattice of a weak del Pezzo surface.
        We consider root subsystems of the root system with Dynkin 
        type either:
            A1, A1+A2, A4, D5, E6, E7 or E8,
        corresponding to ranks 3, 4, 5, 6, 7, 8 and 9 respectively 
        (eg. A1+A2 if rank equals 4, and E8 if rank equals 9).
        Note that root systems live in a subspace of 
        the vector space associated to the Neron-Severi lattice 
        of a weak Del Pezzo surface.        
        
                
        Parameters
        ----------
        max_rank : int
            An integer in [3,...,9].           
    
        Returns
        -------
        list<DPLattice>
            A list of "DPLattice" objects dpl such that dpl.Md_lst 
            is the bases of a root subsystem and dpl.type == A0. 
            The list contains exactly one representative for 
            root subsystems up to equivalence, so that the root
            subsystem defines a unimodular involution.  
        '''
        # check cache
        key = 'get_cls_invo_' + str( rank )
        if key in NSTools.get_tool_dct():
            return NSTools.get_tool_dct()[key]

        bas_lst = DPLattice.get_cls_root_bases( rank )
        counter = 0
        total = len( bas_lst )
        inv_lst = []
        for bas in bas_lst:
            NSTools.p( 'counter =', counter , '/', total, ', root base = ', bas.d_lst )
            counter += 1
            M = basis_to_involution( bas.d_lst, rank )
            if not is_integral_involution( M ):
                continue
            inv = DPLattice( [], bas.d_lst, M )
            inv.set_attributes()
            inv_lst += [ inv ]

        # store in cache
        inv_lst.sort()
        NSTools.get_tool_dct()[key] = inv_lst
        NSTools.save_tool_dct()

        return inv_lst


    @staticmethod
    def get_cls_real_dp( rank = 7 ):
        '''
        Parameters
        ----------
        max_rank : int
            An integer in [3,...,9].           
    
        Returns
        -------
        list<DPLattice>
            A list of DPLattice objects corresponding to Neron-Severi lattices 
            of weak Del Pezzo surfaces of degree (10-rank). The list contains
            exactly one representative for each equivalence class.
              
            All the Div objects referenced in the DPLattice objects of 
            the output have the default intersection matrix:
                diagonal matrix with diagonal: (1,-1,...,-1). 
        '''
        # check cache
        key = 'get_cls_real_dp_' + str( rank )
        if key in NSTools.get_tool_dct():
            return NSTools.get_tool_dct()[key]

        inv_lst = DPLattice.get_cls_involutions( rank )
        bas_lst = DPLattice.get_cls_root_bases( rank )


        # ETA
        ostart = time.time()
        ototal = len( bas_lst ) * len( inv_lst )
        ocounter = 0
        icounter = 0
        ival = 20
        total_orbit_len = 0


        # we fix an involution up to equivalence and go through
        # all possible root bases for singularities.
        dpl_lst = []
        for inv in inv_lst:
            for bas in bas_lst:

                orbit_lst = get_root_bases_orbit( bas.d_lst )

                # ETA outer loop
                ocounter += 1
                otime = ( time.time() - ostart ) / 60
                total_orbit_len += len( orbit_lst )
                avg_orbit_len = total_orbit_len / ocounter
                itotal = ( avg_orbit_len * ototal )
                NSTools.p( '#orbit =', len( orbit_lst ),
                           '(average =', avg_orbit_len, '),',
                           'ocounter =', ocounter, '/', ototal, ',',
                           'rank =', rank,
                           'total minutes passed =', otime,
                           'ETA in minutes =', ( otime / ocounter ) * ( ototal - ocounter ) )


                for d_lst in orbit_lst:

                    # ETA inner loop
                    if icounter % ival == 0:
                        istart = time.time()
                    icounter += 1
                    if icounter % ival == 0:
                        itime = ( time.time() - istart ) / 60
                        NSTools.p( 'ETA in minutes =', ( itime / ival ) * ( itotal - icounter ), ',',
                                   'icounter =', icounter, '/', itotal )

                    # check whether involution inv.M preserves d_lst
                    dm_lst = [ d.mat_mul( inv.M ) for d in d_lst ]
                    dm_lst.sort()
                    if dm_lst != d_lst:
                        continue

                    # add to classification if not equivalent to objects
                    # in list, see "DPLattice.__eq__()".
                    dpl = DPLattice( d_lst, inv.Md_lst, inv.M )
                    if dpl not in dpl_lst:
                        dpl.set_attributes()
                        dpl_lst += [dpl]


        # store in cache
        dpl_lst.sort()
        NSTools.get_tool_dct()[key] = dpl_lst
        NSTools.save_tool_dct()

        return dpl_lst


    @staticmethod
    def get_reduced_cls( rank = 9, invo_cls = False ):
        '''
        Parameters
        ----------
        rank : int
        invo_cls : bool
        
        Returns
        -------
        Classification of DPLattices using weak equivalence. 
        If rank>7, then either dpl.type or dpl.Mtype equals A0.
        
        If "invo_cls==True", then only return DPLattices dpl 
        such that dpl.type is A0.   
        '''
        # check cache
        key = 'get_reduced_cls' + str( ( rank, invo_cls ) )
        if key in NSTools.get_tool_dct():
            return NSTools.get_tool_dct()[key]

        # obtain nonreduced classification
        dpl_lst = []
        if invo_cls:
            dpl_lst = DPLattice.get_cls_involutions( rank )
        else:
            if rank > 7:
                # In this case we do not have the complete
                # classification yet: either type or Mtype equals A0
                dpl_lst += DPLattice.get_cls_root_bases( rank )
                dpl_lst += DPLattice.get_cls_involutions( rank )
            else:
                dpl_lst = DPLattice.get_cls_real_dp( rank )

        # reduce classification
        DPLattice.weak_equality = True  # see DPLattice.__eq__()
        out_lst = []
        for dpl in dpl_lst:

            if not dpl in out_lst:
                out_lst += [dpl]
            else:
                # dpl is already in list using weak equality
                for i in range( len( out_lst ) ):
                    if out_lst[i] == dpl and out_lst[i].Mtype < dpl.Mtype:
                        out_lst[i] = dpl
                        break
        DPLattice.weak_equality = False

        # store in cache
        NSTools.get_tool_dct()[key] = out_lst
        NSTools.save_tool_dct()

        return out_lst


    # overloading of "=="
    # returns True if isomorphic as Neron-Severi lattices
    def __eq__( self, other ):

        # compared with None?
        if type( self ) != type( other ):
            return False

        # cardinality of classes agree?
        if len( self.d_lst ) != len( other.d_lst ):
            return False
        self.set_attributes( 0 )
        other.set_attributes( 0 )
        if len( self.m1_lst ) != len( other.m1_lst ):
            return False
        self.set_attributes( 1 )
        other.set_attributes( 1 )
        if len( self.fam_lst ) != len( other.fam_lst ):
            return False
        self.set_attributes( 2 )
        other.set_attributes( 2 )
        if len( self.real_d_lst ) != len( other.real_d_lst ):
            return False
        self.set_attributes( 3 )
        other.set_attributes( 3 )
        if len( self.real_m1_lst ) != len( other.real_m1_lst ):
            return False
        self.set_attributes( 4 )
        other.set_attributes( 4 )
        if len( self.real_fam_lst ) != len( other.real_fam_lst ):
            return False
        self.set_attributes( 5 )
        other.set_attributes( 5 )
        if len( self.or_lst ) != len( other.or_lst ):
            return False
        self.set_attributes( 6 )
        other.set_attributes( 6 )
        if len( self.sr_lst ) != len( other.sr_lst ):
            return False

        # Dynkin type effective (-2)-classes agree?
        self.set_attributes( 7 )
        other.set_attributes( 7 )
        if self.type != other.type:
            return False

        # Dynkin type real structures agree?
        self.set_attributes( 8 )
        other.set_attributes( 8 )
        if not DPLattice.weak_equality:
            # In a finer classification, this check should be
            # omitted, since equivalent real structures may have
            # different Dynkin types.
            if self.Mtype != other.Mtype:
                return False

        # check Cremona invariant
        self.set_attributes( 9 )
        other.set_attributes( 9 )
        if not self.G.is_isomorphic( other.G, edge_labels = True ):
            if not DPLattice.weak_equality:
                NSTools.p( 'Non isomorphic graphs (unexpected position): ', self, other )
            return False

        return True


    # operator overloading for <
    # Used for sorting lists of DPLattice objects:
    #     <http://stackoverflow.com/questions/1227121/compare-object-instances-for-equality-by-their-attributes-in-python>
    def __lt__( self, other ):

        if self.get_rank() != other.get_rank():
           return self.get_rank() < other.get_rank()

        if len( self.Md_lst ) != len( other.Md_lst ):
           return len( self.Md_lst ) < len( other.Md_lst )

        if len( self.d_lst ) != len( other.d_lst ):
           return len( self.d_lst ) < len( other.d_lst )

        self.set_attributes( 8 )
        other.set_attributes( 8 )

        if self.Mtype != other.Mtype:
            return self.type < other.type

        if self.type != other.type:
            return self.type < other.type

        if len( self.real_fam_lst ) != len( other.real_fam_lst ):
            return len( self.real_fam_lst ) > len( other.real_fam_lst )

        if len( self.fam_lst ) != len( other.fam_lst ):
            return len( self.fam_lst ) < len( other.fam_lst )

        if len( self.m1_lst ) != len( other.m1_lst ):
            return len( self.m1_lst ) < len( other.m1_lst )


    # overloading of "str()": human readable string representation of object
    def __str__( self ):

        self.set_attributes()

        s = '\n'
        s += 50 * '=' + '\n'

        s += 'Degree          = ' + str( self.get_degree() ) + '\n'
        s += 'Rank            = ' + str( self.get_rank() ) + '\n'
        s += 'Intersection    = ' + str( list( self.m1_lst[0].int_mat ) ) + '\n'
        s += 'Real structure  = ' + str( self.Mtype ) + '\n'
        s += 'Singularities   = ' + str( self.type ) + '\n'
        s += 'Cardinalities   = ' + '(' + str( len( self.or_lst ) ) + ', ' + str( len( self.sr_lst ) ) + ')\n'

        arrow = '  --->  '

        s += 'Real involution:\n'
        b_lst = [Div( row ) for row in sage_identity_matrix( sage_ZZ, self.get_rank() ).rows() ]
        for b in b_lst:
            s += '\t' + str( b ) + arrow + str( b.mat_mul( self.M ) ) + '\n'

        s += 'Indecomposable (-2)-classes:\n'
        for d in self.d_lst:
            s += '\t' + str( d ) + arrow + str( d.mat_mul( self.M ) ) + '\n'
        s += '\t#real = ' + str( len( self.real_d_lst ) ) + '\n'

        s += 'Indecomposable (-1)-classes:\n'
        for m1 in self.m1_lst:
            s += '\t' + str( m1 ) + arrow + str( m1.mat_mul( self.M ) ) + '\n'
        s += '\t#real = ' + str( len( self.real_m1_lst ) ) + '\n'

        s += 'Classes of conical families:\n'
        for fam in self.fam_lst:
            s += '\t' + str( fam ) + arrow + str( fam.mat_mul( self.M ) ) + '\n'
        s += '\t#real = ' + str( len( self.real_fam_lst ) ) + '\n'

        s += 50 * '=' + '\n'

        return s





'''
Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
Created on Aug 11, 2016
@author: Niels Lubbes

Algorithm for computing elements in a unimodular lattice.
We use this algorithm in the context of Neron-Severi lattices
of weak del Pezzo surfaces.

See 
    Arxiv: "Computing curves on real rational surfaces"

'''

import time

from ns_lattice.sage_interface import sage_Combinations
from ns_lattice.sage_interface import sage_Compositions
from ns_lattice.sage_interface import sage_Partitions
from ns_lattice.sage_interface import sage_ZZ
from ns_lattice.sage_interface import sage_Permutations

from ns_lattice.class_ns_tools import NSTools
from ns_lattice.class_div import Div


def get_divs( d, dc, cc, perm = False ):
    '''
    Computes divisors in unimodular lattice with prescribed intersection product. 
    
    Parameters
    ---------- 
    d : Div    
        object d0*e0 + d1*e1 +...+ dr*er such that 
            * product signature equals (1,d.rank()-1)
            * d0>0
            * d1,...,dr<=0                         
    dc : int 
        A positive integer.
    
    cc : int        
        Self intersection.
        
    perm : boolean
        If True, then generators are permuted.      
    
    Returns
    -------
    list<Div> 
    
        Returns a sorted list of "Div" objects
        
            * c = c0*e0 + c1*e1 +...+ cr*er
          
        such that 
                        
            * d.rank() == r+1
            * dc       == d*c   (signature = (1,rank-1))
            * cc       == c*c   (signature = (1,rank-1))           
        
        and each Div object satisfies exactly one
        of the following conditions:    
        
            * c == ei - ej   for 0>i>j>=r,
            * c == ei        for i>0, or          
            * c0 > 0,   c1,...,cr <= 0
                           
        If "perm" is False, then then only one representative
        for each c is returned up to permutation of ei for i>0. 
        For example, e0-e1-e2 and e0-e1-e3 are considered equivalent, 
        and only e0-e1-e2 is returned, since e0-e1-e2>e0-e1-e3 
        (see "get_div_set()" for the ordering). In particular,
        c1 >= c2 >= ... >= cr.  
        
    Note
    ----
        If d=[3]+8*[-1], (dc,cc)==(0,-2) and perm=False 
        then the Div classes are 
        '12', '1123', '212' and '308'.
        See "Div.get_label()" for the notation.
        These classes correspond to the (-2)-classes 
        in the Neron-Severi lattice associated to 
        a weak del Pezzo surface.
        
        If perm==False then only one representative
        for each q is returned up to permutation of 
        ei for i>0. For example, e0-e1-e2 and e0-e1-e3
        are considered equivalent, and only e0-e1-e2
        is returned, since e0-e1-e2>e0-e1-e3 
        (see "Div.__lt__()" for the ordering).             
    '''

    # check if input was already computed
    #
    key = 'get_divs_' + str( ( d, dc, cc, perm ) )
    if key in NSTools.get_tool_dct():
        return NSTools.get_tool_dct()[key]

    # construct div set
    #
    NSTools.p( 'Constructing div set classes for ', ( d, dc, cc, perm ) )
    out_lst = []

    # compute classes of the form ei or ei-ej for i,j>0
    #
    if ( dc, cc ) == ( 1, -1 ) or ( dc, cc ) == ( 0, -2 ):

        m2_lst = []  # list of divisors of the form ei-ej for i,j>0
        m1_lst = []  # list of divisors of the form ei for i>0
        if perm:

            # Example:
            #     >>> list(Combinations( [1,2,3,4], 2 ))
            #     [[1, 2], [1, 3], [1, 4], [2, 3], [2, 4], [3, 4]]
            # Notice that r=d.rank()-1 if c = c0*e0 + c1*e1 +...+ cr*er.
            #
            for comb in sage_Combinations( range( 1, d.rank() ), 2 ):
                m2_lst += [ Div.new( str( comb[0] ) + str( comb[1] ), d.rank() ) ]
            m1_lst += [Div.new( 'e' + str( i ), d.rank() ) for i in range( 1, d.rank() )]

        else:

            # up to permutation of the generators
            # we may assume that i==1 and j==2.
            #
            m2_lst += [ Div.new( '12', d.rank() ) ]
            m1_lst += [ Div.new( 'e1', d.rank() ) ]

        # add the classes that satisfy return
        # specification to the output list
        #
        for c in m1_lst + m2_lst:
            if  ( dc, cc ) == ( d * c, c * c ):
                out_lst += [c]

    #
    # Note: cc = c0^2 - c1^2 -...- cr^2
    #
    c0 = 0
    cur_eq_diff = -1
    while True:

        c0 = c0 + 1
        dc_tail = d[0] * c0 - dc  #    = d1*c1 +...+ dr*cr
        dd_tail = d[0] ** 2 - d * d  # = d1^2  +...+ dr^2
        cc_tail = c0 ** 2 - cc  #      = c1^2  +...+ cr^2

        # not possible according to io-specs.
        #
        if dc_tail < 0 or dd_tail < 0 or cc_tail < 0:
            NSTools.p( 'continue... (c0, dc_tail, dd_tail, cc_tail) =', ( c0, dc_tail, dd_tail, cc_tail ) )
            if dd_tail < 0:
                raise Exception( 'dd_tail =', dd_tail )
            continue

        # Cauchy-Schwarz inequality <x,y>^2 <= <x,x>*<y,y> holds?
        #
        prv_eq_diff = cur_eq_diff
        cur_eq_diff = abs( dc_tail * dc_tail - dd_tail * cc_tail )
        if prv_eq_diff == -1:
            prv_eq_diff = cur_eq_diff

        NSTools.p( 'prv_eq_diff =', prv_eq_diff, ', cur_eq_diff =', cur_eq_diff, ', dc_tail^2 =', dc_tail * dc_tail, ', dd_tail*cc_tail =', dd_tail * cc_tail, ', (c0, dc_tail, dd_tail, cc_tail) =', ( c0, dc_tail, dd_tail, cc_tail ) )

        if prv_eq_diff < cur_eq_diff and dc_tail * dc_tail > dd_tail * cc_tail:
            NSTools.p( 'stop by Cauchy-Schwarz inequality...' )
            break  # out of while loop

        # obtain all possible [d1*c1+1,...,dr*cr+1]
        #
        r = d.rank() - 1
        if perm and len( set( d[1:] ) ) != 1:
            p_lst_lst = sage_Compositions( dc_tail + r, length = r )
        else:
            p_lst_lst = sage_Partitions( dc_tail + r, length = r )

        # data for ETA computation
        total = len( p_lst_lst )
        counter = 0
        ival = 5000

        # obtain [c1,...,cr] from [d1*c1+1,...,dr*cr+1]
        #
        for p_lst in p_lst_lst:

            # ETA
            if counter % ival == 0:
                start = time.time()
            counter += 1
            if counter % ival == 0:
                passed_time = time.time() - start
                NSTools.p( 'ETA in minutes =', passed_time * ( total - counter ) / ( ival * 60 ), ' (', counter, '/', total, '), c0 =', c0, ', prv_eq_diff =', prv_eq_diff, ', cur_eq_diff =', cur_eq_diff )

            # dc_tail=d1*c1 +...+ dr*cr = p1 +...+ pr  with pi>=0
            p_lst = [ p - 1 for p in p_lst]

            # obtain c_tail=[c1,...,cr] from [p1,...,pr]
            valid_part = True
            c_tail = []  # =[c1,...,cr]
            for i in range( 0, len( p_lst ) ):
                if p_lst[i] == 0 or d[i + 1] == 0:
                    c_tail += [p_lst[i]]
                else:
                    quo, rem = sage_ZZ( p_lst[i] ).quo_rem( d[i + 1] )
                    if rem != 0:
                        valid_part = False
                        break  # out of i-for-loop
                    else:
                        c_tail += [ quo ]
            if not valid_part:
                continue

            # add to out list if valid
            #
            c = Div( [c0] + c_tail )
            if c.rank() == d.rank() and ( dc, cc ) == ( d * c, c * c ):
                if perm and len( set( d[1:] ) ) == 1:
                    # since d1==...==dr we do not have to
                    # check each permutation.
                    for pc_tail in sage_Permutations( c_tail ):
                        out_lst += [Div( [c0] + list( pc_tail ) )]
                else:
                    out_lst += [c]

    # sort list of "Div" objects
    out_lst.sort()


    # cache output
    NSTools.get_tool_dct()[key] = out_lst
    NSTools.save_tool_dct()

    return out_lst


def get_indecomp_divs( c_lst, d_lst ):
    '''
    Parameters
    ----------
    c_lst : list<Div>        
        Typically output of "get_divs(...)"
    d_lst : list<Div>
        Typically a list of (-2)-classes.

    Returns
    -------
    list<Div>
        Returns a list of "Div" objects c in c_lst, 
        so that c*d >= 0 for all d in "d_lst".                
        
    Note
    ----
    If the Div object represent effective divisor classes in 
    a the Neron-Severi lattice of a weak del Pezzo 
    surface and if d_lst are the classes of singularities,
    then the output correspond to "indecomposable" classes.
    Such classes cannot be written as the sum of effective 
    divisors.    
    '''

    # check positivity against "d_lst"
    out_lst = []
    for c in c_lst:
        indecomp = True
        for d in d_lst:
            if d * c < 0:
                indecomp = False
                break  # out of for loop
        if indecomp:
            out_lst += [c]

    return out_lst


def get_ak( rank ):
    '''
    Parameters
    ----------
    rank : int
    
    Returns
    -------
    Div
        A Div object of given rank of the form
        
            3e0 - e1 - ... - er
        
        Mathematically this is the anticanonical 
        class of the blowup of the projective plane.    
    '''
    return Div( [3] + ( rank - 1 ) * [-1] )




'''
Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
Created on Jan 27, 2018
@author: Niels Lubbes
'''

from ns_lattice.sage_interface import sage_n

from ns_lattice.class_ns_tools import NSTools

import time


class ETA( object ):
    '''
    For estimation when a loop in a program terminates.
    During the loop feedback is printed.  
    '''

    def __init__( self, total, ival ):
        '''

        Should be called before a loop in the program starts.        
        
        Parameters
        ----------
        total: int
            Number of times the loop needs to be traced.
        
        ival : int
            Number of traced loops in program until
            feedback about etimated end time is printed        
        '''
        self.ostart = time.time()  # start time
        self.total = total  # total number of loops
        self.ival = ival  # number of loops after which eta is updated
        self.counter = 0  # loop counter
        self.istart = None  # start time which is updated after ival loops.
        self.eta = None  # estimated time of arrival in minutes


    def update( self, info ):
        '''
        Prints estimation for when program terminates
        (ETA=estimated time of arrival). 
        Should be called inside a loop.  
        
        Parameters
        ----------
        info : str
            A string of additional information that
            is printed together with ETA.
        '''
        if self.counter % self.ival == 0:
            self.istart = time.time()

        self.counter += 1

        if self.counter % self.ival == 0:

            itime = ( time.time() - self.istart ) / ( 60 * self.ival )
            otime = sage_n( ( time.time() - self.ostart ) / 60, digits = 5 )
            self.eta = sage_n( itime * ( self.total - self.counter ), digits = 5 )

            NSTools.p( 'ETA =', self.eta, 'm,',
                       'counter =', self.counter, '/', self.total, ',',
                       'time =', otime, 'm',
                       'info =', info )





'''
Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
Created on Aug 11, 2016
@author: Niels Lubbes
'''

import sys
import os

from ns_lattice.sage_interface import sage_matrix
from ns_lattice.sage_interface import sage_ZZ
from ns_lattice.sage_interface import sage_identity_matrix
from ns_lattice.sage_interface import sage_Subsets
from ns_lattice.sage_interface import sage_Permutations
from ns_lattice.sage_interface import sage_Graph
from ns_lattice.sage_interface import sage_gcd
from ns_lattice.sage_interface import sage_factor

from ns_lattice.class_ns_tools import NSTools

from ns_lattice.class_div import Div

from ns_lattice.div_in_lattice import get_divs
from ns_lattice.div_in_lattice import get_ak

from ns_lattice.class_dp_lattice import DPLattice

from ns_lattice.ns_basis import get_bases_lst
from ns_lattice.ns_basis import get_webs
from ns_lattice.ns_basis import contains_perm
from ns_lattice.ns_basis import triples

from ns_lattice.convert_to_tex import table_to_tex

from linear_series.class_poly_ring import PolyRing
from linear_series.class_base_points import BasePointTree
from linear_series.class_linear_series import LinearSeries


def usecase__get_cls_root_bases( max_rank ):
    '''
    Classification of root bases in root system of rank at most "max_rank".
    See "DPLattice.get_cls_root_bases()".
    
    Parameters
    ----------
    max_rank : int
        Maximal rank.  
    '''
    NSTools.p( 'Classification of root bases.' )

    for rank in range( 3, max_rank + 1 ):
        for dpl in DPLattice.get_cls_root_bases( rank ):
            print( str( rank ) + '\t' + dpl.type + '\t\t\t' + str( dpl.d_lst ) )
        print( 'total = ' + str( len( DPLattice.get_cls_root_bases( rank ) ) ) )


def usecase__get_reduced_cls_dp_lattice( max_rank, tex = False ):
    '''
    We output a table where each row corresponds to exactly 
    one isomorphism class of a Neron-Severi lattices of weak 
    del Pezzo surfaces.
          
    See "DPLattice.usecase__get_reduced_cls_dp_lattice()".  
    
    The table headers of the output table mean the following:
        #(-1)  : length of DPLattice.m1_lst
        #R(-1) : length of DPLattice.real_m1_lst 
        #(0)   : length of DPLattice.fam_lst
        #R(0)  : length of DPLattice.real_fam_lst
        P1xP1  : y iff there exists f and g in DPLattice.real_fam_lst such that f*g=1   
    
    Parameters
    ----------
    max_rank : int
        Maximal rank for the root system associated to singularities and
        involutions of weak del Pezzo surfaces.
                    
    tex : bool
        If True, then output the classification as a table in Tex code.
        The following Tex macros are used for the table header:        
            \newcommand{\MmfD}{{\mathfrak{D}}}
            \newcommand{\MbbP}{{\mathbb{P}}}    
            \newcommand{\Mrow}[3]{\ensuremath{#1\colon #2\longrightarrow #3}}             
    '''
    NSTools.p( 'Reduced classification of DPLattice objects.' )

    #
    # table 1
    #
    tab1 = []
    h1_lst = ['', 'deg', 'real', 'sing', 'P1xP1', '#(-1)', '#(0)', '#R(-1)', '#R(0)']
    t1_lst = [ '', '$\deg(Y)$', '$\MmfD(A(Y))$', '$\MmfD(B(Y))$',
                '$\MbbP^1\\times\MbbP^1$', '$\#E_\star(Y)$',
                '$\#F_\star(Y)$', '$\#E_\MbbR(Y)$', '$\#G(Y)$' ]
    idx = 0
    for rank in range( 3, max_rank + 1 ):
        for dpl in DPLattice.get_reduced_cls( rank, False ):
            row = []
            row += [idx]
            row += [dpl.get_degree()]
            row += [dpl.get_marked_Mtype()]
            row += [dpl.type]
            if dpl.contains_fam_pair():
                row += ['y']
            else:
                row += ['n']
            row += dpl.get_numbers()[1:3] + dpl.get_numbers()[4:]

            tab1 += [row]
            idx += 1

    #
    # table 2
    #
    tab2 = []
    h2_lst = ['', 'deg', 'real', 'involution' ]
    t2_lst = ['', '$\deg(Y)$', '$\MmfD(A(Y))$', '$\sigma_*$']
    idx = 0
    for rank in range( 3, max_rank + 1 ):
        for dpl in DPLattice.get_reduced_cls( rank, True ):
            div_tup = tuple( [Div( row ).mat_mul( dpl.M ) for row in sage_identity_matrix( rank ) ] )

            row = []
            row += [ idx ]
            row += [ dpl.get_degree() ]
            row += [ dpl.get_marked_Mtype() ]
            row += [ div_tup ]

            tab2 += [row]
            idx += 1

    #
    # table 3
    #
    tab3 = []
    h3_lst = ['', 'deg', 'real', 'sing', 'root base' ]
    t3_lst = ['', '$\deg(Y)$', '$\MmfD(A(Y))$', '$\MmfD(B(Y))$', '$B(Y)$']
    idx = 0
    for rank in range( 3, max_rank + 1 ):
        for dpl in DPLattice.get_reduced_cls( rank, False ):

            row = []
            row += [idx]
            row += [dpl.get_degree()]
            row += [dpl.get_marked_Mtype()]
            row += [dpl.type]
            row += [[ d for d in dpl.d_lst ]]

            tab3 += [row]
            idx += 1

    #
    # ordering for the tables
    #
    tab_h_lst = [( tab2, h2_lst, t2_lst ), ( tab3, h3_lst, t3_lst ), ( tab1, h1_lst, t1_lst )]

    #
    # output the tables
    #
    if not tex:

        for tab, h_lst, t_lst in tab_h_lst:

            num_cols = len( tab[0] )
            tab = [['------'] * num_cols] + tab
            tab = [h_lst] + tab
            tab = [['------'] * num_cols] + tab
            tab += [['------'] * num_cols]
            row_format = "{:<10}" * num_cols  # each column is 8 chars.

            for row in tab:
                print( row_format.format( *row ) )
            for rank in range( 3, max_rank + 1 ):
                print( 'deg =', 10 - rank, ' #lattice-classes =', len( DPLattice.get_reduced_cls( rank ) ) )

            print( 80 * '=' )

    else:

        #
        # parameters for table_to_tex()
        #
        max_len = 61
        row_num = 9
        replace_dct = {'A':'A_', 'D':'D_', 'E':'E_', 'e':'e_', '[':'\\{', ']':'\\}'}
        col_idx = 1

        s = ''
        s += '\n\n\subsection{Classification of involution $\Mrow{\sigma_*}{N(Y)}{N(Y)}$}'
        s += '\n\label{sec:t1}\n\n'
        s += table_to_tex( t2_lst, tab2, replace_dct, col_idx, max_len, row_num )

        s += '\n\n\subsection{Classification of root base $B(Y)$ of singular locus}'
        s += '\n\label{sec:t2}\n\n'
        s += table_to_tex( t3_lst, tab3, replace_dct, col_idx, max_len, row_num )


        s += '\n\n\subsection{Cardinalities of distinguished subsets of $N(Y)$}'
        s += '\n\label{sec:t3}\n\n'
        s += table_to_tex( t1_lst, tab1, replace_dct, col_idx, max_len, row_num )

        print( s )


def usecase__get_classes_dp1( rank ):
    '''
    Computes classes in the Neron-Severi lattice with
    predefined self-intersection and intersection with the 
    canonical class.
    
    Parameters
    ----------
    rank : int  
    '''

    # canonical class
    d = get_ak( rank )

    # basis change
    a_lst = [ 'e0-e1', 'e0-e2']
    a_lst = [ Div.new( a, rank ) for a in a_lst ]
    m1_lst = get_divs( d, 1, -1, True )
    print( d )
    M = sage_identity_matrix( rank )
    d_lst = []
    d_tup_lst = get_bases_lst( a_lst, M, d_lst, m1_lst, False )
    B = sage_matrix( sage_ZZ, [ dt.e_lst for dt in d_tup_lst[0] ] )

    # list the classes
    for ( dc, cc ) in [( 2, 0 ), ( 1, -1 ), ( 0, -2 ), ( 2, 2 ), ( 2, 4 ), ( 3, 1 )]:
        NSTools.p( '(dc, cc) =', ( dc, cc ) )
        c_lst = get_divs( d, dc, cc, False )
        for c in c_lst:
            NSTools.p( '\t\t', c, '\t\t', c.get_basis_change( B ) )


def usecase__hex_webs( max_rank ):
    '''
    Characterize triples of families of conics on weak del Pezzo surfaces
    that are not orthogonal to any (-1)-curve. 
    
    Parameters
    ----------
    max_rank : int  
    '''

    for rank in range( 5, max_rank + 1 ):
        NSTools.p( 'rank =', rank )
        for dpl in DPLattice.get_cls_root_bases( rank ):

            t_lst = triples( dpl, 1 )
            if t_lst != []:
                NSTools.p( dpl )
                NSTools.p( 'dpl.d_lst  =', dpl.d_lst )
                NSTools.p( 'dpl.M      =', list( dpl.M ) )
                NSTools.p( 'dpl.Md_lst =', dpl.Md_lst )
                NSTools.p( '::: t_lst  =', t_lst )
                NSTools.p( 10 * '=*' )


def usecase__graphs( rank ):
    '''
    M. Meringer: Fast Generation of Regular Graphs and Construction of Cages. Journal of Graph Theory 30, 137-146, 1999.
    '''
    rank = 7
    dpl = DPLattice.get_cls_root_bases( rank )[0]
    f_lst = dpl.real_fam_lst

    G = sage_Graph()
    G.add_vertices( range( len( f_lst ) ) )

    num_lst = []
    ne_lst = []  # each entry denotes the number of outgoing edges of some vertex
    for i in range( len( f_lst ) ):
        ne = 0
        for j in range( len( f_lst ) ):
            if f_lst[i] * f_lst[j] > 1:
                ne += 1
                num = f_lst[i] * f_lst[j]
                G.add_edge( i, j, num )
                if num not in num_lst:
                    num_lst += [num]

        if ne not in ne_lst:
            ne_lst += [ne]

    NSTools.p( 'labels      =', sorted( num_lst ) )  # [2,3,4,5,6,7,8]
    NSTools.p( 'regularity  =', sorted( ne_lst ) )  # E8: [2095]
    NSTools.p( 'vertices    =', len( f_lst ) )  # E8: [2160]


    P = G.graphplot( vertex_size = 1,
                     vertex_labels = True,
                     edge_labels = True,
                     color_by_label = False,
                     layout = 'circular' ).plot()

    P.save( os.environ['OUTPUT_PATH'] + 'graph.png' )

    NSTools.p( '#components =', G.connected_components_number() )


def usecase__construct_surfaces():
    '''
    We construct a surface parametrization and its Neron-Severi lattice. 
     
    Requires the linear_series package.
    '''

    # Blowup of projective plane in 3 colinear points
    # and 2 infinitely near points. The image of the
    # map associated to the linear series is a quartic
    # del Pezzo surface with 5 families of conics.
    # Moreover the surface contains 8 straight lines.
    #
    ring = PolyRing( 'x,y,z', True )
    p1 = ( -1, 0 )
    p2 = ( 0, 0 )
    p3 = ( 1, 0 )
    p4 = ( 0, 1 )
    p5 = ( 2, 0 )
    bp_tree = BasePointTree()
    bp_tree.add( 'z', p1, 1 )
    bp_tree.add( 'z', p2, 1 )
    bp_tree.add( 'z', p3, 1 )
    bp = bp_tree.add( 'z', p4, 1 )
    bp.add( 't', p5, 1 )
    ls = LinearSeries.get( [3], bp_tree )
    NSTools.p( ls.get_bp_tree() )
    NSTools.p( 'implicit equation =\n\t', ls.get_implicit_image() )

    # construct NS-lattice where p1=e1,...,p5=e5
    rank = 6
    d_lst = [ 'e0-e1-e2-e3', 'e4-e5' ]  # basepoint p5 is infinitely near to p4
    Md_lst = []
    M = sage_identity_matrix( 6 )
    d_lst = [ Div.new( d, rank ) for d in d_lst ]
    Md_lst = [ Div.new( Md, rank ) for Md in Md_lst ]
    M = sage_matrix( M )
    dpl = DPLattice( d_lst, Md_lst, M )
    NSTools.p( 'Neron-Severi lattice =', dpl )

    # search representative for the equivalence class in classification
    assert dpl in DPLattice.get_reduced_cls( rank )


def usecase__roman_circles():
    '''
    We compute circles on a Roman surface.
    '''
    # parametrization of the Roman surface
    #
    p_lst = '[ z^2+x^2+y^2, -z*x, -x*y, z*y ]'

    # we consider the stereographic projection from
    #     S^3 = { x in P^4 | -x0^2+x1^2+x2^2+x3^2+x4^2 = 0 }
    # where the center of projection is (1:0:0:0:1):
    #     (x0:x1:x2:x3:x4) |---> (x0-x4:x1:x2:x3)

    # inverse stereographic projection into 3-sphere
    #
    s_lst = '[ y0^2+y1^2+y2^2+y3^2, 2*y0*y1, 2*y0*y2, 2*y0*y3, -y0^2+y1^2+y2^2+y3^2 ]'

    # compose p_lst with s_lst
    #
    ring = PolyRing( 'x,y,z,y0,y1,y2,y3' )
    x, y, z, y0, y1, y2, y3 = ring.gens()
    p_lst = ring.coerce( p_lst )
    s_lst = ring.coerce( s_lst )
    dct = { y0:p_lst[0], y1:p_lst[1], y2:p_lst[2], y3:p_lst[3] }
    sp_lst = [ s.subs( dct ) for s in s_lst ]
    NSTools.p( 'sp_lst =' )
    for sp in sp_lst: NSTools.p( '\t\t', sage_factor( sp ) )
    NSTools.p( 'gcd(sp_lst) =', sage_gcd( sp_lst ) )

    # determine base points
    #
    ring = PolyRing( 'x,y,z', True )
    sp_lst = ring.coerce( sp_lst )
    ls = LinearSeries( sp_lst, ring )
    NSTools.p( ls.get_bp_tree() )

    # We expect that the basepoints come from the intersection
    # of the Roman surface with the absolute conic:
    #    A = { (y0:y1:y2:y3) in P^3 | y0=y1^2+y2^2+y3^2 = 0 }
    #
    # Circles are the image via p_lst of lines that pass through
    # complex conjugate points.
    #
    ring = PolyRing( 'x,y,z', False )  # reinitialize ring with updated numberfield
    a0, a1, a2, a3 = ring.root_gens()

    # a0=(1-I*sqrt(3)) with conjugate a0-1 and minimal polynomial t^2-t+1

    # we compute candidate classes of circles
    #
    h = Div.new( '4e0-e1-e2-e3-e4-e5-e6-e7-e8' )
    div_lst = get_divs( h, 2, -2, False ) + get_divs( h, 2, -1, False )
    NSTools.p( 'Classes of circles up to permutation:' )
    for c in div_lst:
        NSTools.p( '\t\t', c )

    # We recover the preimages of circles in the Roman surface
    # under the map p_lst, by constructing for each candidate
    # class the corresponding linear series.

    # 2e0-e1-e2-e3-e4-e5-e6-e7-e8
    b = [( a0 - 1, -a0 ), ( -a0, a0 - 1 )]
    b += [( -a0 + 1, a0 ), ( a0, -a0 + 1 )]
    b += [ ( a0 - 1, a0 ), ( -a0, -a0 + 1 )]
    b += [( -a0 + 1, -a0 ), ( a0, a0 - 1 )]
    bp_tree = BasePointTree()
    for i in range( 6 ):
        bp_tree.add( 'z', b[i], 1 )
    NSTools.p( 'basepoints =', b )
    NSTools.p( LinearSeries.get( [2], bp_tree ) )

    # e0-e1-e2
    b = [( a0 - 1, -a0 ), ( -a0, a0 - 1 )]
    bp_tree = BasePointTree()
    bp = bp_tree.add( 'z', b[0], 1 )
    bp = bp_tree.add( 'z', b[1] , 1 )
    NSTools.p( 'basepoints =', b )
    NSTools.p( LinearSeries.get( [1], bp_tree ) )

    # e0-e3-e4
    b = [( -a0 + 1, a0 ), ( a0, -a0 + 1 )]
    bp_tree = BasePointTree()
    bp = bp_tree.add( 'z', b[0], 1 )
    bp = bp_tree.add( 'z', b[1] , 1 )
    NSTools.p( 'basepoints =', b )
    NSTools.p( LinearSeries.get( [1], bp_tree ) )

    # e0-e5-e6
    b = [ ( a0 - 1, a0 ), ( -a0, -a0 + 1 )]
    bp_tree = BasePointTree()
    bp = bp_tree.add( 'z', b[0], 1 )
    bp = bp_tree.add( 'z', b[1] , 1 )
    NSTools.p( 'basepoints =', b )
    NSTools.p( LinearSeries.get( [1], bp_tree ) )

    # e0-e7-e8
    b = [( -a0 + 1, -a0 ), ( a0, a0 - 1 )]
    bp_tree = BasePointTree()
    bp = bp_tree.add( 'z', b[0], 1 )
    bp = bp_tree.add( 'z', b[1] , 1 )
    NSTools.p( 'basepoints =', b )
    NSTools.p( LinearSeries.get( [1], bp_tree ) )

    return


if __name__ == '__main__':

    #  Debug output settings
    #
    mod_lst = []
    mod_lst += ['__main__.py']
    NSTools.filter( mod_lst )  # output only from specified modules
    # NSTools.filter( None )  # print all verbose output, comment to disable.
    # NSTools.get_tool_dct().clear()  # uncomment to remove all cache!

    if 'OUTPUT_PATH' not in os.environ:
        os.environ['OUTPUT_PATH'] = './'

    NSTools.start_timer()

    #
    # Should be between 3 and 9.
    # computes classifications up to rank "max_rank".
    #
    rank = 9

    #########################################
    #                                       #
    # (Un)comment one or more use cases     #
    #                                       #
    #########################################

    usecase__get_cls_root_bases( rank )
    usecase__get_reduced_cls_dp_lattice( rank, False )
    usecase__get_classes_dp1( rank )
    usecase__hex_webs( rank )
    usecase__graphs( rank )
    usecase__construct_surfaces()
    # usecase__roman_circles() # takes about 3 minutes

    #########################################
    #                                       #
    # End of list of use case methods.      #
    #                                       #
    #########################################

    NSTools.end_timer()
    NSTools.p( 'The End' )






'''
Use of this source code is governed by a MIT-style license that can be found in the LICENSE file.
Created on Jan 27, 2017
@author: Niels Lubbes

https://python-packaging.readthedocs.io/en/latest/minimal.html
https://pypi.python.org/pypi?%3Aaction=list_classifiers
'''


from setuptools import setup


setup( name = 'ns_lattice',
       version = '0',
       description = 'Algorithms for computing in Neron-Severi lattice',
       classifiers = [
           'Development Status :: 3 - Alpha',
           'License :: OSI Approved :: MIT License',
           'Programming Language :: Python :: 2',
           'Programming Language :: Python :: 3',
           'Topic :: Scientific/Engineering :: Mathematics',
           ],
      keywords = 'Neron-Severi-lattice',
      url = 'http://github.com/niels-lubbes/ns_lattice',
      author = 'Niels Lubbes',
      license = 'MIT',
      package_dir = {'': 'src'},
      packages = ['ns_lattice'],
      install_requires = ['linear_series'],
      test_suite = 'nose.collector',
      tests_require = ['nose'],
      entry_points = {
          'console_scripts': ['run-lattice=ns_lattice.__main__:main'],
      },
      include_package_data = True,
      zip_safe = False )





'''
Created on Aug 11, 2016
@author: Niels

See [http://arxiv.org/abs/1302.6678] for more info.

Classification of root subsystems of root systems
of type either A1, A1+A2, A4, D5, E6, E7 or E8.
'''
import time

from ns_lattice.sage_interface import sage_VectorSpace
from ns_lattice.sage_interface import sage_vector
from ns_lattice.sage_interface import sage_QQ
from ns_lattice.sage_interface import sage_identity_matrix
from ns_lattice.sage_interface import sage_Graph
from ns_lattice.sage_interface import sage_Partitions
from ns_lattice.sage_interface import sage_RootSystem
from ns_lattice.sage_interface import sage_Subsets
from ns_lattice.sage_interface import sage_Combinations
from ns_lattice.sage_interface import sage_Permutations

from ns_lattice.class_ns_tools import NSTools

from ns_lattice.class_div import Div

from ns_lattice.div_in_lattice import get_divs
from ns_lattice.div_in_lattice import get_indecomp_divs
from ns_lattice.div_in_lattice import get_ak


def is_root_basis( d_lst ):
    '''
    Parameters
    ----------
    d_lst : list<Div> 
        A list of lists of "Div" objects "d", 
        such that d*d=-2 and d*(-3h+e1+...+er)=0 
        where r=rank-1 and rank in [3,...,7].
               
    Returns
    -------
    bool
        True if input is the empty list or if divisors 
        in "d_lst" are linear independent as vectors
        and their pairwise product is either -2, 0 or 1.               
    '''

    if d_lst == []:
        return True

    # check pairwise inner product
    for i in range( len( d_lst ) ):
        for j in range( len( d_lst ) ):
            if d_lst[i] * d_lst[j] not in [0, 1, -2]:
                return False

    # check linear independence
    V = sage_VectorSpace( sage_QQ, d_lst[0].rank() )
    W = V.subspace( [d.e_lst for d in d_lst] )
    return W.rank() == len( d_lst )


def get_graph( d_lst ):
    '''
    Parameters
    ----------
    d_lst : list<Div>
        A list of "Div" objects.

    Returns
    -------
    sage_Graph

        A labeled "Graph()" where the elements
        of "d_lst" are the vertices.
        Different vertices are connected if
        their corresponding intersection product
        is non-zero and the edge is labeled with
        the intersection product.
    '''
    G = sage_Graph()
    G.add_vertices( range( len( d_lst ) ) );

    for i in range( len( d_lst ) ):
        for j in range( len( d_lst ) ):
            if d_lst[i] * d_lst[j] > 0 and i != j:
                G.add_edge( i, j, d_lst[i] * d_lst[j] )

    return G


def get_ext_graph( d_lst, M ):
    '''
    Parameters
    ----------
    d_lst : list<Div> 
        A list of "Div" objects of equal rank.
        
    M : sage_matrix<sage_ZZ>   
        A square matrix with integral coefficients
        of rank "d_lst[0].rank()" 
    
    Returns
    -------
        A labeled "sage_Graph()" where the elements 
        of "d_lst" are the vertices. 
        A pair of vertices are connected 
        by and edge labeled with their 
        non-zero intersection product. 
        Two vertices which are related 
        via M are connected with an edge labeled 1000.
        Labeled self-loops are also included.
        Two orthogonal vertices are connected 
    '''
    NSTools.p( 'd_lst =', len( d_lst ), d_lst, ', M =', list( M ) )

    G = sage_Graph()
    G.add_vertices( range( len( d_lst ) ) )

    for i in range( len( d_lst ) ):
        for j in range( len( d_lst ) ):
            if d_lst[i] * d_lst[j] != 0:
                G.add_edge( i, j, d_lst[i] * d_lst[j] )

    for i in range( len( d_lst ) ):
        j = d_lst.index( d_lst[i].mat_mul( M ) )
        G.add_edge( i, j, 1000 )

    return G


def get_dynkin_type( d_lst ):
    '''
    Parameters
    ----------
    d_lst : list<Div>
        A list of lists of "Div" objects "d" of 
        the same rank, such that 
            d*d=-2 and d*(-3h+e1+...+er)=0 
        where 
            r=rank-1 and rank in [3,...,9].  
        We assume that "is_root_basis(d_lst)==True":
        linear independent, self intersection number -2
        and pairwise product either 0 or 1.            
                     
    Returns
    -------
    string
        Returns a string denoting the Dynkin type of a 
        root system with basis "d_lst".  
        Returns 'A0' if "d_lst==[]".
    
    Note
    ----
        For example:
        [<1145>, <1123>, <23>, <45>, <56>, <78>] --> '3A1+A3'
        where <1145> is shorthand for "Div.new('1145')".      
        
    Raises
    ------
    ValueError
        If the Dynkin type of d_lst cannot be recognized.
         
    '''
    if d_lst == []: return 'A0'

    # check whether values are cached
    #
    construct_dynkin_types = True
    max_r = d_lst[0].rank() - 1
    key = 'get_dynkin_type_' + str( max_r )
    for r in range( max_r, 8 + 1 ):
        if 'get_dynkin_type_' + str( r ) in NSTools.get_tool_dct():
            key = 'get_dynkin_type_' + str( r )
            construct_dynkin_types = False

    # construct list of dynkin types if values are not cached
    #
    if construct_dynkin_types:
        NSTools.p( 'Constructing list of Dynkin types... max_r =', max_r )

        ade_lst = []
        for comb_lst in sage_Combinations( max_r * ['A', 'D', 'E'], max_r ):
            for perm_lst in sage_Permutations( comb_lst ):
                ade_lst += [perm_lst]
        #
        # "ade_lst" contains all combinations of 'A', 'D', 'E'
        # and looks as follows:
        #
        #     ade_lst[0] = ['A', 'A', 'A', 'A', 'A', 'A', 'A', 'A']
        #     ade_lst[1] = ['A', 'A', 'A', 'A', 'A', 'A', 'A', 'D']
        #     ade_lst[2] = ['A', 'A', 'A', 'A', 'A', 'A', 'D', 'A']
        #     ...
        #     ade_lst[?] = ['A', 'D', 'A', 'D', 'A', 'D', 'E', 'A']
        #     ...
        #     ade_lst[-1]= ['E', 'E', 'E', 'E', 'E', 'E', 'E', 'E']
        #

        type_lst = []
        ts_lst = []
        for ade in ade_lst:
            for r in range( 1, max_r + 1 ):
                for p_lst in sage_Partitions( r + max_r, length = max_r ):

                    # obtain type list
                    t_lst = [( ade[i], p_lst[i] - 1 ) for i in range( max_r ) if  p_lst[i] != 1]
                    t_lst.sort()

                    # obtain Root system
                    # or continue if invalid Cartan/Dynkin type
                    if ( 'D', 2 ) in t_lst or ( 'D', 3 ) in t_lst:
                        continue
                    try:
                        rs = sage_RootSystem( t_lst )
                    except ValueError as err:
                        continue  # not a valid Cartan type

                    # obtain graph G
                    mat = list( -1 * rs.cartan_matrix() )
                    G = sage_Graph()
                    G.add_vertices( range( len( mat ) ) );
                    for i in range( len( mat ) ):
                        for j in range( len( mat[0] ) ):
                           if mat[i][j] == 1:
                               G.add_edge( i, j )

                    # obtain string for type
                    # Example: [(A,1),(A,1),(A,1),(A,3)] ---> '3A1+A3'
                    tmp_lst = [t for t in t_lst]
                    ts = ''
                    while len( tmp_lst ) > 0:
                        t = tmp_lst[0]
                        c = tmp_lst.count( t )
                        while t in tmp_lst:
                            tmp_lst.remove( t )
                        if ts != '':
                            ts += '+'
                        if c > 1:
                            ts += str( c )

                        ts += t[0] + str( t[1] )

                    # add to type_lst if new
                    if ts not in ts_lst:
                        type_lst += [( G, ts, t_lst )]
                        ts_lst += [ts]
                        NSTools.p( 'added to list: ', ts, '\t\t...please wait...' )

        NSTools.p( 'Finished constructing list of Dynkin types.' )
        # cache the constructed "type_lst"
        NSTools.get_tool_dct()[key] = type_lst
        NSTools.save_tool_dct()

    # end if
    else:
        type_lst = NSTools.get_tool_dct()[key]
    G1 = get_graph( d_lst )

    # loop through all types and check equivalence
    for ( G2, ts, t_lst ) in type_lst:
        if G1.is_isomorphic( G2 ):
            return ts

    raise ValueError( 'Could not recognize Dynkin type: ', d_lst )


def get_root_bases_orbit( d_lst, positive = True ):
    '''
    Computes the orbit of a root base under the Weyl group.
    
    Parameters
    ----------
    d_lst : list<Div>
        A list of lists of "Div" objects "d" of the same rank or the empty list.    

    positive : bool
    
    Returns
    -------
    list<list<Div>>
        A list of distinct lists of "Div" objects "d" of the same rank. 
        such that d*d=-2 and d*(-3h+e1+...+er)=0 where r=rank-1.
        
        If "d_lst" is the empty list, then "[]" is returned.
        
        Otherwise we return a list of root bases such that each root basis
        is obtained as follows from a root "s" such that s*s=-2 
        and s*(-3h+e1+...+er)=0: 
        
            [ d + (d*s)d for d in d_lst ]
        
        We do this for all possible roots in [s1,s2,s3,...]: 
        
            [ [ d + (d*s1)d for d in d_lst ],  [ d + (d*s2)d for d in d_lst ], ... ]
         
        Mathematically, this means that we consider the Weyl group 
        of the root system with Dynkin type determined by the rank of elements 
        in "d_lst". The Dynkin type is either 
            A1, A1+A2, A4, D5, E6, E7 or E8.
        We return the orbit of the elements in "d_lst" under
        the action of the Weyl group.
                
        If "positive==True" then the roots in the basis are all positive
        and thus of the form 
            <ij>, <1ijk>, <2ij>, <30i>
        with i<j<k. 
        For example '15' and '1124' but not '-15' or '-1124'. 
        See "Div.get_label()" for the notation.                           
    '''
    if d_lst == []:
        return [[]]

    rank = d_lst[0].rank()

    # in cache?
    key = 'get_root_bases_orbit_' + str( d_lst ) + '_' + str( rank )
    if key in NSTools.get_tool_dct():
        return NSTools.get_tool_dct()[key]

    # obtain list of all positive (-2)-classes
    m2_lst = get_divs( get_ak( rank ), 0, -2, True )
    # m2_lst += [ m2.int_mul( -1 ) for m2 in m2_lst]
    NSTools.p( 'd_lst  =', len( d_lst ), d_lst, ', m2_lst =', len( m2_lst ), m2_lst )

    # data for ETA computation
    counter = 0
    total = len( m2_lst )
    ival = 5000

    d_lst.sort()
    d_lst_lst = [d_lst]
    for cd_lst in d_lst_lst:

        total = len( m2_lst ) * len( d_lst_lst )
        for m2 in m2_lst:

            # ETA
            if counter % ival == 0:
                start = time.time()
            counter += 1
            if counter % ival == 0:
                passed_time = time.time() - start
                NSTools.p( 'ETA in minutes =', passed_time * ( total - counter ) / ( ival * 60 ), ', len(d_lst_lst) =', len( d_lst_lst ), ', total =', total )

            #
            # The action of roots on a root base is by reflection:
            #     cd - 2(cd*m2/m2*m2)m2
            # Notice that m2*m2==-2.
            #
            od_lst = [ cd + m2.int_mul( cd * m2 ) for cd in cd_lst]
            # print( 'm2 =', m2, ', od_lst =', od_lst, ', cd_lst =', cd_lst, ', d_lst_lst =', d_lst_lst, ' positive =', positive )

            od_lst.sort()
            if od_lst not in d_lst_lst:
                d_lst_lst += [od_lst]

    # select positive roots if positive==True
    pd_lst_lst = []
    for d_lst in d_lst_lst:
        if positive and '-' in [ d.get_label( True )[0] for d in d_lst ]:
            continue  # continue with for loop since a negative root in basis
        pd_lst_lst += [d_lst]

    # cache output
    NSTools.get_tool_dct()[key] = pd_lst_lst
    NSTools.save_tool_dct()

    NSTools.p( '#orbit(' + str( d_lst ) + ') =', len( pd_lst_lst ) )

    return pd_lst_lst




'''
Created on Aug 11, 2016
@author: Niels Lubbes

Classification of unimodular involutions of Neron-Severi lattice 
of weak del Pezzo surfaces.
'''

from ns_lattice.sage_interface import sage_ZZ
from ns_lattice.sage_interface import sage_QQ
from ns_lattice.sage_interface import sage_matrix
from ns_lattice.sage_interface import sage_identity_matrix
from ns_lattice.sage_interface import sage_diagonal_matrix

from ns_lattice.class_ns_tools import NSTools

from ns_lattice.class_div import Div

from ns_lattice.div_in_lattice import get_ak


def complete_basis( d_lst ):
    '''
    Parameters
    ----------
    d_lst : list<Div>
        A list of "Div" objects.
    
    Returns
    -------
    sage_matrix<sage_QQ>
        Returns a square matrix over QQ of full rank. The first columns
        correspond to the elements in d_lst (where d_lst is sorted). 
        The appended columns are orthogonal to the first "len(d_lst)" columns.
         
    Examples
    --------
        We explain with 3 examples where the dotted (:) vertical lines 
        denote which columns are appended. 
          
          | 0  0 : 1  0  0  0 | | 0  0  0 : 1  0  0 | | 0  0  0  1 : -3  0 |  
          | 0  0 : 0 -1  0  0 | | 0  0  0 : 0 -1  0 | | 1  0  0 -1 :  1  0 |
          | 0  0 : 0  0 -1  0 | | 1  0  0 : 0  0 -1 | |-1  1  0 -1 :  1  0 |
          | 1  0 : 0  0  0 -1 | |-1  1  0 : 0  0 -1 | | 0 -1  0 -1 :  1  0 |
          |-1  1 : 0  0  0 -1 | | 0 -1  1 : 0  0 -1 | | 0  0  1  0 :  0  1 |
          | 0 -1 : 0  0  0 -1 | | 0  0 -1 : 0  0 -1 | | 0  0 -1  0 :  0  1 |
             
    '''
    # sort
    d_lst = [ d for d in d_lst]
    d_lst.sort()

    # extend with orthogonal vectors
    row_lst = [ d.e_lst for d in d_lst]
    ext_lst = []
    for v_lst in  sage_matrix( sage_ZZ, row_lst ).right_kernel().basis():
        ext_lst += [ [-v_lst[0]] + list( v_lst[1:] ) ]  # accounts for signature (1,rank-1)
    mat = sage_matrix( sage_QQ, row_lst + ext_lst ).transpose()

    # verify output
    if mat.rank() < d_lst[0].rank():
        raise Error( 'Matrix expected to have full rank: ', d_lst, '\n' + str( mat ) )
    de_lst = [ Div( ext ) for ext in ext_lst ]
    for de in de_lst:
        for d in d_lst:
            if d * de != 0:
                raise Error( 'Extended columns are expected to be orthogonal: ', de, d, de_lst, d_lst, list( mat ) )

    return mat


def is_integral_involution( M ):
    '''
    Parameters
    ----------
    M : sage_matrix
        A matrix M.
    
    Returns
    -------
    bool
        Returns True if the matrix is an involution, 
        preserves inner product with signature (1,r) 
        and has integral coefficients.
    '''

    nrows, ncols = M.dimensions()

    # check whether involution
    if M * M != sage_identity_matrix( nrows ):
        return False

    # check whether inner product is preserved
    S = sage_diagonal_matrix( [1] + ( ncols - 1 ) * [-1] )
    if M.transpose() * S * M != S:
        return False

    # check whether coefficients are integral
    for r in range( nrows ):
        for c in range( ncols ):
            if M[r][c] not in sage_ZZ:
                return False

    # check whether canonical class is preserved
    ak = get_ak( nrows )
    if ak.mat_mul( M ) != ak:
        return False

    return True


def basis_to_involution( d_lst, rank ):
    '''
    Parameters
    ----------
    d_lst : list<Div>
        A list of "Div" objects of rank "rank".
    
    rank : int
        An integer in [3,...,9].
    
    Returns
    -------
    sage_MATRIX<sage_QQ>
        Returns matrix over QQ that correspond to an involution of 
            ZZ<h,e1,...,er>
        here r=rank-1. The first columns
        correspond to the elements in d_lst (where d_lst is sorted). 
        The appended columns are orthogonal to the first "len(d_lst)" columns.    
    '''
    if d_lst == []:
        return sage_identity_matrix( sage_QQ, rank )

    l = len( d_lst )
    V = complete_basis( d_lst )
    D = sage_diagonal_matrix( l * [-1] + ( rank - l ) * [1] )
    M = V * D * V.inverse()  # MV=VD

    return M

